import copy
from dataclasses import dataclass, field
from datetime import datetime
import pathlib
import re
from typing import List, Optional, Tuple, Union
import uuid

from common import (
    MULTI_LINE_REQUIREMENTS,
    OPENAPI_NO_VALIDATION,
    RequirementsFile,
    SINGLE_LINE_REQUIREMENTS,
    TEST_COMPUTE_CONFIG_DICT,
    TEST_CONTAINERFILE,
    TEST_REQUIREMENTS_FILES,
    TestLogger,
)
import pytest

from anyscale import Anyscale
from anyscale._private.anyscale_client import (
    FakeAnyscaleClient,
    WORKSPACE_CLUSTER_NAME_PREFIX,
)
from anyscale._private.models.image_uri import ImageURI
from anyscale._private.sdk.timer import FakeTimer
from anyscale.client.openapi_client.models import (
    Cloud,
    ComputeTemplate,
    ComputeTemplateConfig,
    CreateInternalProductionJob,
    HaJobGoalStates,
    HaJobStates,
    InternalProductionJob,
    ProductionJobConfig,
    ProductionJobStateTransition,
)
from anyscale.client.openapi_client.models.create_job_queue_config import (
    CreateJobQueueConfig,
)
from anyscale.client.openapi_client.models.job_queue_spec import JobQueueSpec
from anyscale.client.openapi_client.models.job_status import (
    JobStatus as BackendJobStatus,
)
from anyscale.client.openapi_client.models.production_job import ProductionJob
from anyscale.client.openapi_client.models.ray_runtime_env_config import (
    RayRuntimeEnvConfig,
)
from anyscale.job import JobSDK
from anyscale.job.models import (
    JobConfig,
    JobLogMode,
    JobQueueConfig,
    JobQueueExecutionMode,
    JobQueueSpec as JobQueueSpecSDK,
    JobRunState,
    JobRunStatus,
    JobState,
    JobStatus,
)
from anyscale.sdk.anyscale_client.models.job import Job as APIJobRun
from anyscale.shared_anyscale_utils.latest_ray_version import LATEST_RAY_VERSION


@pytest.fixture()
def job_sdk_with_fakes(
    sdk_with_fakes: Tuple[Anyscale, FakeAnyscaleClient, TestLogger, FakeTimer]
) -> Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer]:
    sdk, client, logger, timer = sdk_with_fakes
    return sdk.job, client, logger, timer


class TestSubmit:
    def test_basic(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        config = JobConfig(entrypoint="python hello.py", name="test-job-name",)
        sdk.submit(config)
        assert fake_client.submitted_job == CreateInternalProductionJob(
            name="test-job-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            config=ProductionJobConfig(
                entrypoint="python hello.py",
                runtime_env=None,
                max_retries=1,
                build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
                compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ),
        )

    def test_default_name(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        config = JobConfig(entrypoint="python hello.py",)
        sdk.submit(config)
        assert fake_client.submitted_job.name
        assert fake_client.submitted_job == CreateInternalProductionJob(
            name=fake_client.submitted_job.name,
            project_id=fake_client.DEFAULT_PROJECT_ID,
            config=ProductionJobConfig(
                entrypoint="python hello.py",
                runtime_env=None,
                max_retries=1,
                build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
                compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ),
        )

    def test_custom_image_from_containerfile(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        fake_containerfile = pathlib.Path(TEST_CONTAINERFILE).read_text()
        job_name = "test-job-name"
        build_id = fake_client.get_cluster_env_build_id_from_containerfile(
            cluster_env_name=f"image-for-job-{job_name}",
            containerfile=fake_containerfile,
            anonymous=False,
        )

        config = JobConfig(
            entrypoint="python hello.py",
            name=job_name,
            containerfile=TEST_CONTAINERFILE,
        )
        sdk.submit(config)
        assert fake_client.submitted_job == CreateInternalProductionJob(
            name=job_name,
            project_id=fake_client.DEFAULT_PROJECT_ID,
            config=ProductionJobConfig(
                entrypoint="python hello.py",
                runtime_env=None,
                max_retries=1,
                build_id=build_id,
                compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ),
        )

    def test_custom_image_from_image_uri(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        job_name = "test-job-name"
        image_uri = ImageURI.from_str("docker.io/user/my-custom-image:latest")
        build_id = fake_client.get_cluster_env_build_id_from_image_uri(
            image_uri=image_uri, ray_version="2.23.0",
        )

        config = JobConfig(
            entrypoint="python hello.py",
            name=job_name,
            image_uri=str(image_uri),
            ray_version="2.23.0",
        )
        sdk.submit(config)
        assert fake_client.submitted_job == CreateInternalProductionJob(
            name=job_name,
            project_id=fake_client.DEFAULT_PROJECT_ID,
            config=ProductionJobConfig(
                entrypoint="python hello.py",
                runtime_env=None,
                max_retries=1,
                build_id=build_id,
                compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ),
        )
        assert fake_client.submitted_job
        build_id = fake_client.submitted_job.config.build_id
        build = fake_client.get_cluster_env_build(build_id)
        assert build
        assert build.ray_version == "2.23.0"

        sdk.submit(
            config.options(image_uri="docker.io/user/does-not-exist:latest").options(
                ray_version=None
            )
        )
        assert fake_client.submitted_job
        new_build_id = fake_client.submitted_job.config.build_id
        new_build = fake_client.get_cluster_env_build(new_build_id)
        assert new_build
        assert new_build.docker_image_name == "docker.io/user/does-not-exist:latest"
        assert new_build.ray_version == LATEST_RAY_VERSION

    def test_custom_compute_config_name(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        fake_client.add_compute_config(
            ComputeTemplate(
                id="compute_id123",
                name="my-custom-compute-config",
                config=ComputeTemplateConfig(
                    cloud_id="test-cloud-id",
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config = JobConfig(
            entrypoint="python hello.py",
            name="test-job-name",
            compute_config="my-custom-compute-config",
        )
        sdk.submit(config)
        assert fake_client.submitted_job == CreateInternalProductionJob(
            name="test-job-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            config=ProductionJobConfig(
                entrypoint="python hello.py",
                runtime_env=None,
                max_retries=1,
                build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
                compute_config_id="compute_id123",
            ),
        )

        with pytest.raises(
            ValueError, match="The compute config 'does-not-exist' does not exist."
        ):
            sdk.submit(config.options(compute_config="does-not-exist"))

    def test_custom_compute_config_dict(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        fake_client.add_cloud(
            Cloud(
                id=str(uuid.uuid4()),
                name=TEST_COMPUTE_CONFIG_DICT["cloud"],
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
        )

        # Specify a compute config as a dictionary (anonymous).
        config = JobConfig(
            entrypoint="python hello.py",
            name="test-job-name",
            compute_config=TEST_COMPUTE_CONFIG_DICT,
        )
        sdk.submit(config)
        anonymous_compute_config_id = fake_client.submitted_job.config.compute_config_id
        anonymous_compute_config = fake_client.get_compute_config(
            anonymous_compute_config_id
        )
        assert anonymous_compute_config.anonymous
        assert anonymous_compute_config.id == anonymous_compute_config_id
        assert fake_client.submitted_job == CreateInternalProductionJob(
            name="test-job-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            config=ProductionJobConfig(
                entrypoint="python hello.py",
                runtime_env=None,
                max_retries=1,
                build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
                compute_config_id=anonymous_compute_config_id,
            ),
        )

        # Test an invalid dict.
        with pytest.raises(
            TypeError,
            match=re.escape("__init__() got an unexpected keyword argument 'bad'"),
        ):
            sdk.submit(
                JobConfig(
                    entrypoint="python hello.py",
                    name="test-job-name",
                    compute_config={"bad": "config"},
                )
            )

    def test_max_retries(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        config = JobConfig(
            entrypoint="python hello.py", name="test-job-name", max_retries=10,
        )
        sdk.submit(config)
        assert fake_client.submitted_job == CreateInternalProductionJob(
            name="test-job-name",
            project_id=fake_client.DEFAULT_PROJECT_ID,
            config=ProductionJobConfig(
                entrypoint="python hello.py",
                runtime_env=None,
                max_retries=10,
                build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
                compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ),
        )

    def test_project(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        project = "test-project-name"
        project_id = fake_client.register_project_by_name(project)
        config = JobConfig(
            entrypoint="python hello.py", name="test-job-name", project=project,
        )
        sdk.submit(config)
        assert fake_client.submitted_job == CreateInternalProductionJob(
            name="test-job-name",
            project_id=project_id,
            config=ProductionJobConfig(
                entrypoint="python hello.py",
                runtime_env=None,
                max_retries=1,
                build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
                compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
            ),
        )

        with pytest.raises(
            ValueError,
            match="Cloud does not match from provided `cloud` and `compute_config`. Either pass one of these or ensure they match.",
        ):
            fake_client.add_cloud(
                Cloud(
                    id=str(uuid.uuid4()),
                    name=TEST_COMPUTE_CONFIG_DICT["cloud"],
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
            )
            config = JobConfig(
                entrypoint="python hello.py",
                name="test-job-name",
                compute_config=TEST_COMPUTE_CONFIG_DICT,
                cloud=fake_client.DEFAULT_CLOUD_NAME,
            )
            sdk.submit(config)

    @pytest.mark.parametrize("target_compute_config", [None, "existing-compute-config"])
    def test_job_queue_creation(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
        target_compute_config: Optional[str],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        target_compute_config_id = None

        if target_compute_config:
            target_compute_config_id = "cpt_123"

            fake_client.add_compute_config(
                ComputeTemplate(
                    id=target_compute_config_id,
                    name=target_compute_config,
                    config=ComputeTemplateConfig(
                        cloud_id="test-cloud-id",
                        local_vars_configuration=OPENAPI_NO_VALIDATION,
                    ),
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                )
            )

        job_queue_config = JobQueueConfig(
            priority=100,
            job_queue_spec=JobQueueSpecSDK(
                name="test-job-queue",
                execution_mode=JobQueueExecutionMode.PRIORITY,  # type: ignore
                compute_config=target_compute_config,
                max_concurrency=100,
                idle_timeout_s=3600,
            ),
        )

        config = JobConfig(
            entrypoint="python hello.py",
            name="test-job-name",
            job_queue_config=job_queue_config,
            compute_config=target_compute_config,
        )

        sdk.submit(config)

        assert (
            fake_client.submitted_job.to_dict()
            == CreateInternalProductionJob(
                name="test-job-name",
                project_id="fake-default-project-id",
                config=ProductionJobConfig(
                    entrypoint="python hello.py",
                    runtime_env=None,
                    max_retries=1,
                    build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
                    compute_config_id=target_compute_config_id
                    or fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
                ),
                job_queue_config=CreateJobQueueConfig(
                    priority=100,
                    job_queue_spec=JobQueueSpec(
                        job_queue_name="test-job-queue",
                        execution_mode=JobQueueExecutionMode.PRIORITY,
                        compute_config_id=target_compute_config_id,
                        max_concurrency=100,
                        idle_timeout_sec=3600,
                    ),
                ),
            ).to_dict()
        )

    def test_job_queue_association(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        job_queue_config = JobQueueConfig(
            priority=100, target_job_queue_name="test-priority-queue",
        )

        config = JobConfig(
            entrypoint="python hello.py",
            name="test-job-name",
            job_queue_config=job_queue_config,
        )

        sdk.submit(config)

        assert (
            fake_client.submitted_job.to_dict()
            == CreateInternalProductionJob(
                name="test-job-name",
                project_id="fake-default-project-id",
                config=ProductionJobConfig(
                    entrypoint="python hello.py",
                    runtime_env=None,
                    max_retries=1,
                    build_id=fake_client.DEFAULT_CLUSTER_ENV_BUILD_ID,
                    compute_config_id=fake_client.DEFAULT_CLUSTER_COMPUTE_ID,
                ),
                job_queue_config=CreateJobQueueConfig(
                    priority=100, target_job_queue_name="test-priority-queue",
                ),
            ).to_dict()
        )


class TestDeployWorkspaceDefaults:
    def test_name_defaults_to_workspace_name(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        # Happy path: workspace cluster name has the expected prefix.
        fake_client.set_inside_workspace(
            True,
            cluster_name=WORKSPACE_CLUSTER_NAME_PREFIX + "super-special-workspace-name",
        )
        sdk.submit(JobConfig(entrypoint="python hello.py"))
        assert fake_client.submitted_job.name == "job-super-special-workspace-name"

        # Defensive path: workspace cluster name doesn't have the expected prefix.
        fake_client.set_inside_workspace(
            True, cluster_name="not-sure-how-this-happened"
        )
        sdk.submit(JobConfig(entrypoint="python hello.py"))
        assert fake_client.submitted_job.name == "job-not-sure-how-this-happened"

    def test_pick_up_cluster_configs(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        fake_client.set_inside_workspace(True)

        fake_client.add_compute_config(
            ComputeTemplate(
                id=fake_client.WORKSPACE_CLUSTER_COMPUTE_ID,
                config=ComputeTemplateConfig(
                    cloud_id=fake_client.WORKSPACE_CLOUD_ID,
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        sdk.submit(
            JobConfig(
                entrypoint="python hello.py",
                name="test-job-name",
                working_dir="s3://remote.zip",
            )
        )
        assert fake_client.submitted_job == CreateInternalProductionJob(
            name="test-job-name",
            project_id=fake_client.WORKSPACE_PROJECT_ID,
            workspace_id=fake_client.WORKSPACE_ID,
            config=ProductionJobConfig(
                entrypoint="python hello.py",
                runtime_env={"working_dir": "s3://remote.zip"},
                max_retries=1,
                build_id=fake_client.WORKSPACE_CLUSTER_ENV_BUILD_ID,
                compute_config_id=fake_client.WORKSPACE_CLUSTER_COMPUTE_ID,
            ),
        )


class TestOverrideApplicationRuntimeEnvs:
    @pytest.mark.parametrize(
        "new_py_modules", [None, [], ["A"], ["C", "D"]],
    )
    def test_update_py_modules(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
        new_py_modules: Optional[List[str]],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        job = JobConfig(
            name="test-job-name",
            entrypoint="python hello.py",
            py_modules=new_py_modules,
        )

        sdk.submit(job)
        submitted_runtime_env = fake_client.submitted_job.config.runtime_env
        if new_py_modules:
            uploaded_py_modules = [
                fake_client._upload_uri_mapping[local_dir]
                for local_dir in new_py_modules
            ]
        else:
            uploaded_py_modules = None

        if uploaded_py_modules:
            assert submitted_runtime_env.get("py_modules") == uploaded_py_modules
        else:
            assert (
                submitted_runtime_env is None
                or "py_modules" not in submitted_runtime_env
            )

    @pytest.mark.parametrize("excludes_override", [None, ["override"]])
    @pytest.mark.parametrize(
        "working_dir_override", [None, "./some-local-path", "s3://some-remote-path.zip"]
    )
    @pytest.mark.parametrize("inside_workspace", [False, True])
    def test_override_working_dir_excludes(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
        inside_workspace: bool,
        working_dir_override: Optional[str],
        excludes_override: Optional[List[str]],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        fake_client.set_inside_workspace(inside_workspace)

        job = JobConfig(
            name="test-job-name",
            entrypoint="python hello.py",
            working_dir=working_dir_override,
            excludes=excludes_override,
        )

        sdk.submit(job)
        submitted_runtime_env = fake_client.submitted_job.config.runtime_env

        cloud_id = (
            FakeAnyscaleClient.WORKSPACE_CLOUD_ID
            if inside_workspace
            else FakeAnyscaleClient.DEFAULT_CLOUD_ID
        )
        cwd_uri = fake_client.upload_local_dir_to_cloud_storage(".", cloud_id=cloud_id)
        if working_dir_override is None and not inside_workspace:
            assert (
                submitted_runtime_env is None
                or "working_dir" not in submitted_runtime_env
            )
        elif working_dir_override is None and inside_workspace:
            assert submitted_runtime_env["working_dir"] == cwd_uri
        elif working_dir_override is not None and working_dir_override.startswith("s3"):
            assert submitted_runtime_env["working_dir"] == working_dir_override
        else:
            fake_client.upload_local_dir_to_cloud_storage(
                "./some-local-path", cloud_id=cloud_id
            )

        if excludes_override is None:
            assert (
                submitted_runtime_env is None or "excludes" not in submitted_runtime_env
            )
        else:
            assert submitted_runtime_env["excludes"] == ["override"]

    @pytest.mark.parametrize(
        "requirements_override",
        [None, MULTI_LINE_REQUIREMENTS.get_path(), ["override"]],
    )
    @pytest.mark.parametrize("workspace_tracking_enabled", [False, True])
    @pytest.mark.parametrize(
        "enable_image_build_for_tracked_requirements", [False, True]
    )
    @pytest.mark.parametrize("inside_workspace", [False, True])
    def test_override_requirements(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
        inside_workspace: bool,
        workspace_tracking_enabled: bool,
        enable_image_build_for_tracked_requirements: bool,
        requirements_override: Union[None, str, List[str]],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        sdk._private_sdk._image_sdk._enable_image_build_for_tracked_requirements = (
            enable_image_build_for_tracked_requirements
        )
        fake_client.set_inside_workspace(
            inside_workspace,
            requirements_path=SINGLE_LINE_REQUIREMENTS.get_path()
            if workspace_tracking_enabled
            else None,
        )

        job = JobConfig(
            name="test-job-name",
            entrypoint="python hello.py",
            requirements=requirements_override,
        )

        sdk.submit(job)
        if (
            inside_workspace
            and workspace_tracking_enabled
            and enable_image_build_for_tracked_requirements
        ):
            build_id = fake_client.submitted_job.config.build_id
            assert build_id
            build = fake_client.get_cluster_env_build(build_id)
            assert build.containerfile == (
                "# syntax=docker/dockerfile:1\n"
                "FROM docker.io/my/base-ws-image:latest\n"
                'RUN pip install "pip-install-test"'
            )
        submitted_runtime_env = fake_client.submitted_job.config.runtime_env

        if isinstance(requirements_override, str):
            # Override with a file.
            pass
        elif isinstance(requirements_override, list):
            # Override with a list.
            pass
        elif (
            inside_workspace
            and workspace_tracking_enabled
            and not enable_image_build_for_tracked_requirements
        ):
            # Workspace default.
            expected_workspace_pip = SINGLE_LINE_REQUIREMENTS.expected_pip_list
            assert submitted_runtime_env["pip"] == expected_workspace_pip
        else:
            # No overrides.
            assert submitted_runtime_env is None or "pip" not in submitted_runtime_env


class TestDeployUploadDirs:
    @pytest.mark.parametrize("inside_workspace", [False, True])
    def test_upload_basic_working_dir(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
        inside_workspace: bool,
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        fake_client.set_inside_workspace(inside_workspace)

        config = JobConfig(entrypoint="python test.py", working_dir=".",)
        original_config = copy.deepcopy(config)

        sdk.submit(config)
        # The original config should not be modified.
        assert config == original_config

        # Check that the correct cloud_id was used for the upload.
        expected_cloud_id = (
            FakeAnyscaleClient.WORKSPACE_CLOUD_ID
            if inside_workspace
            else FakeAnyscaleClient.DEFAULT_CLOUD_ID
        )
        assert fake_client.submitted_job.config.runtime_env["working_dir"].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(cloud_id=expected_cloud_id)
        )

    def test_upload_uses_cloud_from_compute_config(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        fake_client.add_compute_config(
            ComputeTemplate(
                id="fake-compute-config-id",
                name="fake-compute-config",
                config=ComputeTemplateConfig(
                    cloud_id="compute-config-cloud-id",
                    local_vars_configuration=OPENAPI_NO_VALIDATION,
                ),
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )
        )

        config = JobConfig(
            entrypoint="python test.py",
            working_dir=".",
            compute_config="fake-compute-config",
        )
        sdk.submit(config)

        # Check that the correct cloud_id was used for the upload.
        expected_cloud_id = "compute-config-cloud-id"
        assert fake_client.submitted_job.config.runtime_env["working_dir"].startswith(
            FakeAnyscaleClient.CLOUD_BUCKET.format(cloud_id=expected_cloud_id)
        )

    def test_upload_with_no_local_dirs(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        """Configs should be left unchanged if there are no local dirs."""
        sdk, fake_client, _, _ = job_sdk_with_fakes

        basic_config = JobConfig(entrypoint="python test.py",)
        sdk.submit(basic_config)
        assert fake_client.submitted_job.config.runtime_env is None

        config_with_requirements = JobConfig(
            entrypoint="python test.py", requirements=["pip-install-test"],
        )
        sdk.submit(config_with_requirements)
        assert fake_client.submitted_job.config.runtime_env == {
            "pip": ["pip-install-test"]
        }

    def test_no_upload_remote_working_dir(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        config = JobConfig(
            entrypoint="python test.py", working_dir="s3://some-remote-uri.zip",
        )

        sdk.submit(config)
        assert (
            fake_client.submitted_job.config.runtime_env["working_dir"]
            == "s3://some-remote-uri.zip"
        )


class TestLoadRequirementsFiles:
    @pytest.mark.parametrize("requirements", [None, *TEST_REQUIREMENTS_FILES])
    def test_override_requirements_file(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
        requirements: Optional[RequirementsFile],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        job = JobConfig(entrypoint="python hello.py")
        if requirements is not None:
            job = job.options(requirements=requirements.get_path())

        if requirements is not None and requirements.expected_pip_list is None:
            with pytest.raises(FileNotFoundError):
                sdk.submit(job)

            return
        else:
            sdk.submit(job)

        if requirements is None:
            assert fake_client.submitted_job.config.runtime_env is None
        else:
            assert (
                fake_client.submitted_job.config.runtime_env["pip"]
                == requirements.expected_pip_list
            )


TEST_JOB_ID = "test-job-id"
TEST_JOB_NAME = "test-job-name"
TEST_COMPUTE_CONFIG_NAME = "fake-compute-config"
OTHER_COMPUTE_CONFIG_NAME = "other-compute-config"
TEST_COMPUTE_CONFIG_ID = "fake-compute-config-id"


DEFAULT_PROD_COMPUTE_CONFIG = ComputeTemplate(
    id=FakeAnyscaleClient.DEFAULT_CLUSTER_COMPUTE_ID,
    name=TEST_COMPUTE_CONFIG_NAME,
    config=ComputeTemplateConfig(
        cloud_id="compute-config-cloud-id",
        local_vars_configuration=OPENAPI_NO_VALIDATION,
    ),
    local_vars_configuration=OPENAPI_NO_VALIDATION,
)


OTHER_PROD_COMPUTE_CONFIG = ComputeTemplate(
    id=FakeAnyscaleClient.DEFAULT_CLUSTER_COMPUTE_ID,
    name=OTHER_COMPUTE_CONFIG_NAME,
    config=ComputeTemplateConfig(
        cloud_id="other-compute-config-cloud-id",
        local_vars_configuration=OPENAPI_NO_VALIDATION,
    ),
    local_vars_configuration=OPENAPI_NO_VALIDATION,
)

DEFAULT_PROD_JOB_CONFIG = ProductionJobConfig(
    entrypoint="python hello.py",
    runtime_env=None,
    max_retries=1,
    build_id=FakeAnyscaleClient.DEFAULT_CLUSTER_ENV_BUILD_ID,
    compute_config_id=FakeAnyscaleClient.DEFAULT_CLUSTER_COMPUTE_ID,
)


DEFAULT_EXPECTED_JOB_CONFIG = JobConfig(
    name=TEST_JOB_NAME,
    entrypoint="python hello.py",
    compute_config=TEST_COMPUTE_CONFIG_NAME + ":1",
)

DEFAULT_EXPECTED_STATUS = JobStatus(
    id=TEST_JOB_ID,
    name=TEST_JOB_NAME,
    state=JobState.UNKNOWN,
    config=DEFAULT_EXPECTED_JOB_CONFIG,
    runs=[],
)

DEFAULT_JOB_RUN = APIJobRun(
    id="job-run-id",
    ray_session_name="job-run-ray-session-id",
    ray_job_id="job-run-ray-job-id",
    status=BackendJobStatus.SUCCEEDED,
    created_at=datetime.now(),
    cluster_id="job-run-cluster-id",
    runtime_environment_id="job-run-runtime-environment-id",
    creator_id="job-run-creator-id",
    name="job-run-name",
)

DEFAULT_JOB_RUN_STATUS = JobRunStatus(
    name=DEFAULT_JOB_RUN.name, state=JobRunState.SUCCEEDED
)


def _expected_status(**kwargs) -> JobStatus:
    return JobStatus(
        id=kwargs["id"] if "id" in kwargs else DEFAULT_EXPECTED_STATUS.id,
        name=kwargs["name"] if "name" in kwargs else DEFAULT_EXPECTED_STATUS.name,
        state=kwargs["state"] if "state" in kwargs else DEFAULT_EXPECTED_STATUS.state,
        config=kwargs["config"]
        if "config" in kwargs
        else DEFAULT_EXPECTED_STATUS.config,
        runs=kwargs["runs"] if "runs" in kwargs else DEFAULT_EXPECTED_STATUS.runs,
    )


def _job_run(**kwargs) -> APIJobRun:
    return APIJobRun(
        id=kwargs["id"] if "id" in kwargs else DEFAULT_JOB_RUN.id,
        ray_session_name=kwargs["ray_session_name"]
        if "ray_session_name" in kwargs
        else DEFAULT_JOB_RUN.ray_session_name,
        ray_job_id=kwargs["ray_job_id"]
        if "ray_job_id" in kwargs
        else DEFAULT_JOB_RUN.ray_job_id,
        status=kwargs["status"] if "status" in kwargs else DEFAULT_JOB_RUN.status,
        created_at=kwargs["created_at"]
        if "created_at" in kwargs
        else DEFAULT_JOB_RUN.created_at,
        cluster_id=kwargs["cluster_id"]
        if "cluster_id" in kwargs
        else DEFAULT_JOB_RUN.cluster_id,
        runtime_environment_id=kwargs["runtime_environment_id"]
        if "runtime_environment_id" in kwargs
        else DEFAULT_JOB_RUN.runtime_environment_id,
        creator_id=kwargs["creator_id"]
        if "creator_id" in kwargs
        else DEFAULT_JOB_RUN.creator_id,
        name=kwargs["name"] if "name" in kwargs else DEFAULT_JOB_RUN.name,
    )


def _job_run_status(**kwargs) -> JobRunStatus:
    return JobRunStatus(
        name=kwargs["name"] if "name" in kwargs else DEFAULT_JOB_RUN_STATUS.name,
        state=kwargs["state"] if "state" in kwargs else DEFAULT_JOB_RUN_STATUS.state,
    )


@dataclass
class JobStatusTestCase:
    id: str = TEST_JOB_ID
    name: str = TEST_JOB_NAME
    state: Optional[HaJobStates] = None
    compute_config: ComputeTemplate = DEFAULT_PROD_COMPUTE_CONFIG
    prod_job_config: ProductionJobConfig = DEFAULT_PROD_JOB_CONFIG
    expected_status: JobStatus = DEFAULT_EXPECTED_STATUS
    runs: List[APIJobRun] = field(default_factory=list)

    def job_model(self) -> ProductionJob:
        return InternalProductionJob(
            id=self.id,
            name=self.name,
            config=self.prod_job_config,
            state=ProductionJobStateTransition(
                current_state=self.state,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )


JOB_STATUS_TEST_CASES = [
    JobStatusTestCase(),
    JobStatusTestCase(
        state=HaJobStates.UPDATING,
        expected_status=_expected_status(state=JobState.RUNNING,),
    ),
    JobStatusTestCase(
        state=HaJobStates.RUNNING,
        expected_status=_expected_status(state=JobState.RUNNING,),
    ),
    JobStatusTestCase(
        state=HaJobStates.RESTARTING,
        expected_status=_expected_status(state=JobState.RUNNING,),
    ),
    JobStatusTestCase(
        state=HaJobStates.CLEANING_UP,
        expected_status=_expected_status(state=JobState.RUNNING,),
    ),
    JobStatusTestCase(
        state=HaJobStates.PENDING,
        expected_status=_expected_status(state=JobState.STARTING,),
    ),
    JobStatusTestCase(
        state=HaJobStates.AWAITING_CLUSTER_START,
        expected_status=_expected_status(state=JobState.STARTING),
    ),
    JobStatusTestCase(
        state=HaJobStates.SUCCESS,
        expected_status=_expected_status(state=JobState.SUCCEEDED,),
    ),
    JobStatusTestCase(
        state=HaJobStates.ERRORED,
        expected_status=_expected_status(state=JobState.FAILED,),
    ),
    JobStatusTestCase(
        state=HaJobStates.TERMINATED,
        expected_status=_expected_status(state=JobState.FAILED,),
    ),
    JobStatusTestCase(
        state=HaJobStates.BROKEN,
        expected_status=_expected_status(state=JobState.FAILED,),
    ),
    JobStatusTestCase(
        state=HaJobStates.OUT_OF_RETRIES,
        expected_status=_expected_status(state=JobState.FAILED),
    ),
    JobStatusTestCase(
        state=HaJobStates.SUCCESS,
        prod_job_config=ProductionJobConfig(
            entrypoint="python hello.py",
            runtime_env=RayRuntimeEnvConfig(
                working_dir="/usr/my_working_dir",
                pip=["package1", "package2"],
                env_vars={"ENV_VAR": "ENV_VAR_VALUE"},
            ),
            max_retries=1,
            build_id=FakeAnyscaleClient.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config_id=FakeAnyscaleClient.DEFAULT_CLUSTER_COMPUTE_ID,
        ),
        expected_status=JobStatus(
            id=TEST_JOB_ID,
            name=TEST_JOB_NAME,
            state=JobState.SUCCEEDED,
            config=JobConfig(
                name=TEST_JOB_NAME,
                entrypoint="python hello.py",
                working_dir="/usr/my_working_dir",
                requirements=["package1", "package2"],
                env_vars={"ENV_VAR": "ENV_VAR_VALUE"},
                compute_config=TEST_COMPUTE_CONFIG_NAME + ":1",
            ),
            runs=[],
        ),
    ),
    JobStatusTestCase(
        state=HaJobStates.SUCCESS,
        compute_config=OTHER_PROD_COMPUTE_CONFIG,
        prod_job_config=ProductionJobConfig(
            entrypoint="python hello.py",
            max_retries=1,
            build_id=FakeAnyscaleClient.DEFAULT_CLUSTER_ENV_BUILD_ID,
            compute_config=OTHER_PROD_COMPUTE_CONFIG,
            compute_config_id=FakeAnyscaleClient.DEFAULT_CLUSTER_COMPUTE_ID,
        ),
        expected_status=JobStatus(
            id=TEST_JOB_ID,
            name=TEST_JOB_NAME,
            state=JobState.SUCCEEDED,
            config=JobConfig(
                name=TEST_JOB_NAME,
                entrypoint="python hello.py",
                compute_config=OTHER_COMPUTE_CONFIG_NAME + ":1",
            ),
            runs=[],
        ),
    ),
    JobStatusTestCase(
        state=HaJobStates.SUCCESS,
        runs=[_job_run()],
        expected_status=_expected_status(
            state=JobState.SUCCEEDED, runs=[_job_run_status()]
        ),
    ),
    JobStatusTestCase(
        state=HaJobStates.ERRORED,
        runs=[_job_run(status=BackendJobStatus.FAILED)],
        expected_status=_expected_status(
            state=JobState.FAILED, runs=[_job_run_status(state=JobRunState.FAILED)]
        ),
    ),
    JobStatusTestCase(
        state=HaJobStates.SUCCESS,
        runs=[
            _job_run(status=BackendJobStatus.FAILED, name="failed-job-run"),
            _job_run(status=BackendJobStatus.SUCCEEDED, name="succeeded-job-run"),
        ],
        expected_status=_expected_status(
            state=JobState.SUCCEEDED,
            runs=[
                _job_run_status(state=JobRunState.FAILED, name="failed-job-run"),
                _job_run_status(state=JobRunState.SUCCEEDED, name="succeeded-job-run"),
            ],
        ),
    ),
]


class TestStatus:
    def test_job_not_found(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, _, _, _ = job_sdk_with_fakes

        with pytest.raises(
            RuntimeError, match="Job with name 'test-job-name' was not found."
        ):
            sdk.status(name="test-job-name")

    @pytest.mark.parametrize("test_case", JOB_STATUS_TEST_CASES)
    def test_build_status_from_model(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
        test_case: JobStatusTestCase,
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        fake_client.add_compute_config(test_case.compute_config)

        prod_job: ProductionJob = test_case.job_model()
        fake_client.update_job(prod_job)

        for run in test_case.runs:
            fake_client.update_job_run(prod_job_id=prod_job.id, model=run)

        status: JobStatus = sdk.status(name=prod_job.name)
        assert status == test_case.expected_status

        status = sdk.status(id=prod_job.id)
        assert status == test_case.expected_status

    def test_status_with_projects(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes

        project = "test-project-name"
        name = "test-job-name"
        cloud = "fake-default-cloud"
        fake_client.register_project_by_name(name=project, cloud="fake-default-cloud")

        fake_client.add_compute_config(DEFAULT_PROD_COMPUTE_CONFIG)

        job_id = sdk.submit(
            JobConfig(
                entrypoint="python hello.py", name=name, cloud=cloud, project=project,
            )
        )

        # submit another job with jobthe same name, under the default project
        sdk.submit(JobConfig(entrypoint="python hello.py", cloud=cloud, name=name))

        # ensure that the project filter works
        status: JobStatus = sdk.status(name=name, cloud=cloud, project=project)
        assert status.id == job_id
        assert status.config.project == project


class TestTerminate:
    def test_not_found(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, _, _, _ = job_sdk_with_fakes

        with pytest.raises(
            RuntimeError, match="Job with name 'does-not-exist' was not found"
        ):
            sdk.terminate(name="does-not-exist")

        with pytest.raises(
            RuntimeError, match="Job with ID 'does-not-exist' was not found"
        ):
            sdk.terminate(id="does-not-exist")

    @pytest.mark.parametrize("terminate_by_name", [True, False])
    def test_basic(
        self,
        terminate_by_name,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        job_id = "test-terminate-job-id"
        name = "test-terminate-job-name"
        prod_job: ProductionJob = InternalProductionJob(
            id=job_id,
            name=name,
            config=DEFAULT_PROD_JOB_CONFIG,
            state=ProductionJobStateTransition(
                current_state=HaJobStates.RUNNING,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )
        fake_client.update_job(prod_job)

        if terminate_by_name:
            sdk.terminate(name=name)
        else:
            sdk.terminate(id=job_id)

        job: Optional[ProductionJob] = fake_client.get_job(
            job_id=job_id, name=None, cloud=None, project=None
        )
        assert job is not None and job.state == HaJobStates.TERMINATED


class TestArchive:
    def test_not_found(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, _, _, _ = job_sdk_with_fakes

        with pytest.raises(
            RuntimeError, match="Job with name 'does-not-exist' was not found"
        ):
            sdk.archive(name="does-not-exist")

        with pytest.raises(
            RuntimeError, match="Job with ID 'does-not-exist' was not found"
        ):
            sdk.archive(id="does-not-exist")

    @pytest.mark.parametrize("archive_by_name", [True, False])
    def test_basic(
        self,
        archive_by_name,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        job_id = "test-terminate-job-id"
        name = "test-terminate-job-name"
        prod_job: ProductionJob = InternalProductionJob(
            id=job_id,
            name=name,
            config=DEFAULT_PROD_JOB_CONFIG,
            state=ProductionJobStateTransition(
                current_state=HaJobStates.SUCCESS,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )
        fake_client.update_job(prod_job)

        assert not fake_client.is_archived_job(job_id)

        if archive_by_name:
            sdk.archive(name=name)
        else:
            sdk.archive(id=job_id)

        assert fake_client.is_archived_job(job_id)


class TestWait:
    def test_job_not_found(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, _, _, _ = job_sdk_with_fakes
        with pytest.raises(
            RuntimeError, match="Job with name 'test-job-name' was not found."
        ):
            sdk.wait(
                name="test-job-name", state=JobState.SUCCEEDED, timeout_s=60,
            )

    def test_succeeds_immediately(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, logger, _ = job_sdk_with_fakes
        job_id = "test-wait-job-id"
        name = "test-wait-job-name"
        prod_job: ProductionJob = InternalProductionJob(
            id=job_id,
            name=name,
            config=DEFAULT_PROD_JOB_CONFIG,
            state=ProductionJobStateTransition(
                current_state=HaJobStates.SUCCESS,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )
        fake_client.update_job(prod_job)

        sdk.wait(
            name=name, state=JobState.SUCCEEDED, timeout_s=60,
        )
        assert any(
            f"Job '{name}' reached target state, exiting" in info_log
            for info_log in logger.info_messages
        )

    def test_times_out(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, logger, _ = job_sdk_with_fakes
        job_id = "test-wait-job-id"
        name = "test-wait-job-name"
        prod_job: ProductionJob = InternalProductionJob(
            id=job_id,
            name=name,
            config=DEFAULT_PROD_JOB_CONFIG,
            state=ProductionJobStateTransition(
                current_state=HaJobStates.RUNNING,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )
        fake_client.update_job(prod_job)

        with pytest.raises(
            TimeoutError,
            match=f"Job '{name}' did not reach target state SUCCEEDED within 60s. Last seen state: RUNNING.",
        ):
            sdk.wait(
                name=name, state=JobState.SUCCEEDED, timeout_s=60,
            )

    def test_job_run_sequence(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, logger, timer = job_sdk_with_fakes

        job_id = "test-wait-job-id"
        name = "test-wait-job-name"
        base_prod_job: ProductionJob = InternalProductionJob(
            id=job_id,
            name=name,
            config=DEFAULT_PROD_JOB_CONFIG,
            state=ProductionJobStateTransition(
                current_state=HaJobStates.PENDING,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            local_vars_configuration=OPENAPI_NO_VALIDATION,
        )
        fake_client.update_job(base_prod_job)

        def state_from_hajobstate(state) -> ProductionJobStateTransition:
            return ProductionJobStateTransition(
                current_state=state,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            )

        def start_sequence(i: int):
            model = copy.deepcopy(base_prod_job)
            if i == 0:
                model.state = state_from_hajobstate(HaJobStates.PENDING)
            elif i == 1:
                model.state = state_from_hajobstate(HaJobStates.RUNNING)
            else:
                model.state = state_from_hajobstate(HaJobStates.SUCCESS)

            fake_client.update_job(model)

        timer.set_on_poll_iteration(start_sequence)

        sdk.wait(
            name=name, state=JobState.SUCCEEDED, timeout_s=60,
        )

        assert any(
            f"Job '{name}' transitioned from STARTING to RUNNING" in info_log
            for info_log in logger.info_messages
        )
        assert any(
            f"Job '{name}' transitioned from RUNNING to SUCCEEDED" in info_log
            for info_log in logger.info_messages
        )
        assert any(
            f"Job '{name}' reached target state, exiting" in info_log
            for info_log in logger.info_messages
        )


class TestLogs:
    def test_not_found(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, _, _, _ = job_sdk_with_fakes

        with pytest.raises(
            RuntimeError, match="Job with name 'does-not-exist' was not found"
        ):
            sdk.get_logs(name="does-not-exist")

        with pytest.raises(
            RuntimeError, match="Job with ID 'does-not-exist' was not found"
        ):
            sdk.get_logs(id="does-not-exist")

    def test_errors_for_max_lines(
        self,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, _, _, _ = job_sdk_with_fakes

        with pytest.raises(TypeError, match="max_lines must be an int"):
            sdk.get_logs(name="name", max_lines="123")

        with pytest.raises(ValueError, match="max_lines must be > 0"):
            sdk.get_logs(name="name", max_lines=-123)

    @pytest.mark.parametrize("get_logs_by_name", [True, False])
    def test_basic(
        self,
        get_logs_by_name,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        job_id = "test-logs-job-id"
        name = "test-logs-job-name"

        failed_run = _job_run(
            id="failed_job_run_id",
            status=BackendJobStatus.FAILED,
            name="failed_job_run_name",
        )
        failed_run_logs = "JOB FAILED"
        fake_client.update_job_run(prod_job_id=job_id, model=failed_run)
        fake_client.add_job_run_logs(failed_run.id, failed_run_logs)

        prod_job: ProductionJob = InternalProductionJob(
            id=job_id,
            name=name,
            config=DEFAULT_PROD_JOB_CONFIG,
            state=ProductionJobStateTransition(
                current_state=HaJobStates.SUCCESS,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            local_vars_configuration=OPENAPI_NO_VALIDATION,
            last_job_run_id=failed_run.id,
        )
        fake_client.update_job(prod_job)

        logs = sdk.get_logs(name=name) if get_logs_by_name else sdk.get_logs(id=job_id)
        assert logs == failed_run_logs

        succeeded_run = _job_run(
            id="succeeded_job_run_id",
            status=BackendJobStatus.SUCCEEDED,
            name="succeeded_job_run_name",
        )
        succeeded_run_logs = "JOB SUCCEEDED"
        fake_client.update_job_run(prod_job_id=prod_job.id, model=succeeded_run)
        fake_client.add_job_run_logs(succeeded_run.id, succeeded_run_logs)

        prod_job.last_job_run_id = succeeded_run.id
        fake_client.update_job(prod_job)

        # without run arg gets latest job run logs
        logs = sdk.get_logs(name=name) if get_logs_by_name else sdk.get_logs(id=job_id)
        assert logs == succeeded_run_logs

        # grab by run name for both job runs
        logs = (
            sdk.get_logs(name=name, run=failed_run.name)
            if get_logs_by_name
            else sdk.get_logs(id=job_id, run=failed_run.name)
        )
        assert logs == failed_run_logs

        logs = (
            sdk.get_logs(name=name, run=succeeded_run.name)
            if get_logs_by_name
            else sdk.get_logs(id=job_id, run=succeeded_run.name)
        )
        assert logs == succeeded_run_logs

    @pytest.mark.parametrize("get_logs_by_name", [True, False])
    def test_head_and_tail_logs(
        self,
        get_logs_by_name: bool,
        job_sdk_with_fakes: Tuple[JobSDK, FakeAnyscaleClient, TestLogger, FakeTimer],
    ):
        sdk, fake_client, _, _ = job_sdk_with_fakes
        job_id = "test-logs-job-id"
        name = "test-logs-job-name"

        run = _job_run(
            id="job_run_id", status=BackendJobStatus.SUCCEEDED, name="job_run_name",
        )
        run_logs_lines = [f"Line {i}" for i in range(100)]
        fake_client.update_job_run(prod_job_id=job_id, model=run)
        fake_client.add_job_run_logs(run.id, "\n".join(run_logs_lines))

        prod_job: ProductionJob = InternalProductionJob(
            id=job_id,
            name=name,
            config=DEFAULT_PROD_JOB_CONFIG,
            state=ProductionJobStateTransition(
                current_state=HaJobStates.SUCCESS,
                goal_state=HaJobGoalStates.SUCCESS,
                local_vars_configuration=OPENAPI_NO_VALIDATION,
            ),
            local_vars_configuration=OPENAPI_NO_VALIDATION,
            last_job_run_id=run.id,
        )
        fake_client.update_job(prod_job)

        # using head to limit the number of logs
        num_lines = 5
        logs = (
            sdk.get_logs(name=name, mode=JobLogMode.HEAD, max_lines=num_lines)
            if get_logs_by_name
            else sdk.get_logs(id=job_id, mode=JobLogMode.HEAD, max_lines=num_lines)
        )
        assert logs == "\n".join(run_logs_lines[:num_lines])

        # using tail to limit the number of logs
        logs = (
            sdk.get_logs(name=name, mode=JobLogMode.TAIL, max_lines=num_lines)
            if get_logs_by_name
            else sdk.get_logs(id=job_id, mode=JobLogMode.TAIL, max_lines=num_lines)
        )
        assert logs == "\n".join(run_logs_lines[-1 * num_lines :])

        # defaults to tailing if mode is not passed
        logs = (
            sdk.get_logs(name=name, max_lines=num_lines)
            if get_logs_by_name
            else sdk.get_logs(id=job_id, max_lines=num_lines)
        )
        assert logs == "\n".join(run_logs_lines[-1 * num_lines :])

        # using strings instead of JobLogMode
        logs = (
            sdk.get_logs(name=name, mode="HEAD", max_lines=num_lines)
            if get_logs_by_name
            else sdk.get_logs(id=job_id, mode="HEAD", max_lines=num_lines)
        )
        assert logs == "\n".join(run_logs_lines[:num_lines])

        logs = (
            sdk.get_logs(name=name, mode="TAIL", max_lines=num_lines)
            if get_logs_by_name
            else sdk.get_logs(id=job_id, mode="TAIL", max_lines=num_lines)
        )
        assert logs == "\n".join(run_logs_lines[-1 * num_lines :])

        # passing max_lines greater than number of lines for both head or tail fetches whole log
        num_lines = 1000
        logs = (
            sdk.get_logs(name=name, mode=JobLogMode.HEAD, max_lines=num_lines)
            if get_logs_by_name
            else sdk.get_logs(id=job_id, mode=JobLogMode.HEAD, max_lines=num_lines)
        )
        assert logs == "\n".join(run_logs_lines)

        logs = (
            sdk.get_logs(name=name, mode=JobLogMode.TAIL, max_lines=num_lines)
            if get_logs_by_name
            else sdk.get_logs(id=job_id, mode=JobLogMode.TAIL, max_lines=num_lines)
        )
        assert logs == "\n".join(run_logs_lines)



from typing import Optional

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale._private.sdk.timer import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import FineTunedModel


class PrivateLLMModelsSDK(BaseSDK):
    def __init__(
        self,
        *,
        logger: Optional[BlockLogger] = None,
        client: Optional[AnyscaleClientInterface] = None,
        timer: Optional[Timer] = None,
    ):
        super().__init__(logger=logger, client=client, timer=timer)

    def list(self):
        raise NotImplementedError

    def retrieve(self, model_id) -> FineTunedModel:
        return self.client.retrieve_finetuned_model(model_id)



import os
from typing import Generator, Optional
from unittest.mock import Mock, patch
import uuid

import click
from click.testing import CliRunner
import pytest

from anyscale._private.sdk import _LAZY_SDK_SINGLETONS
from anyscale.commands.compute_config_commands import (
    archive_compute_config,
    create_compute_config,
    get_compute_config,
)
from anyscale.compute_config.commands import _COMPUTE_CONFIG_SDK_SINGLETON_KEY
from anyscale.compute_config.models import ComputeConfig, ComputeConfigVersion


def _get_test_file_path(subpath: str) -> str:
    return os.path.join(os.path.dirname(__file__), "test_files/", subpath)


MINIMAL_CONFIG_PATH = _get_test_file_path("compute_config_files/minimal.yaml")
FULL_CONFIG_PATH = _get_test_file_path("compute_config_files/full.yaml")

DEFAULT_COMPUTE_CONFIG_ID = "default-fake-compute-config-id"


class FakeComputeConfigSDK:
    DEFAULT_COMPUTE_CONFIG_NAME = "default-fake-compute-config-name"
    DEFAULT_COMPUTE_CONFIG = ComputeConfig(cloud="custom-cloud")

    def __init__(self):
        # Create method.
        self.created_config: Optional[ComputeConfig] = None
        self.created_id: Optional[str] = None
        self.created_name: Optional[str] = None

        # Get method.
        self.fetched_name: Optional[str] = None
        self.fetched_id: Optional[str] = None
        self.fetched_include_archived: Optional[bool] = None

        # Archive method.
        self.archived_name: Optional[str] = None
        self.archived_id: Optional[str] = None

    def create_compute_config(self, config: ComputeConfig, *, name: Optional[str]):
        assert isinstance(config, ComputeConfig)
        self.created_config = config
        self.created_id = str(uuid.uuid4())
        self.created_name = name or self.DEFAULT_COMPUTE_CONFIG_NAME
        return f"{self.created_name}:1", self.created_id

    def get_compute_config(
        self,
        *,
        name: Optional[str],
        id: Optional[str],  # noqa: A002
        include_archived: bool = False,
    ):
        self.fetched_name = name
        self.fetched_id = id
        self.fetched_include_archived = include_archived

        returned_name = name or self.DEFAULT_COMPUTE_CONFIG_NAME
        if ":" not in returned_name:
            returned_name += ":1"
        return ComputeConfigVersion(
            name=returned_name, id=id or str(uuid.uuid4()), config=ComputeConfig(),
        )

    def archive_compute_config(
        self, *, name: Optional[str], id: Optional[str]  # noqa: A002
    ):
        self.archived_name = name
        self.archived_id = id


@pytest.fixture()
def mock_compute_config_controller() -> Generator[Mock, None, None]:
    mock_compute_config_controller = Mock(
        create=Mock(return_value=DEFAULT_COMPUTE_CONFIG_ID),
        get=Mock(return_value={"old": "format", "compute": "config"}),
    )
    mock_compute_config_controller_cls = Mock(
        return_value=mock_compute_config_controller
    )
    with patch(
        "anyscale.commands.compute_config_commands.ComputeConfigController",
        new=mock_compute_config_controller_cls,
    ):
        yield mock_compute_config_controller


@pytest.fixture()
def fake_compute_config_sdk() -> Generator[FakeComputeConfigSDK, None, None]:
    fake_compute_config_sdk = FakeComputeConfigSDK()
    _LAZY_SDK_SINGLETONS[_COMPUTE_CONFIG_SDK_SINGLETON_KEY] = fake_compute_config_sdk
    try:
        yield fake_compute_config_sdk
    finally:
        del _LAZY_SDK_SINGLETONS[_COMPUTE_CONFIG_SDK_SINGLETON_KEY]


def _assert_error_message(result: click.testing.Result, *, message: str):
    assert result.exit_code != 0
    assert message in result.stdout


class TestCreate:
    def test_no_config_file(
        self, fake_compute_config_sdk, mock_compute_config_controller
    ):
        runner = CliRunner()
        result = runner.invoke(create_compute_config)
        _assert_error_message(
            result,
            message="Either the --config-file flag or [COMPUTE_CONFIG_FILE] argument must be provided",
        )

    def test_config_file_argument_and_flag(
        self, fake_compute_config_sdk, mock_compute_config_controller
    ):
        runner = CliRunner()
        result = runner.invoke(
            create_compute_config, ["-f", "flag.yaml", MINIMAL_CONFIG_PATH]
        )
        _assert_error_message(
            result,
            message="Only one of the --config-file flag or [COMPUTE_CONFIG_FILE] argument can be provided",
        )

    def test_config_file_argument_not_found(
        self, fake_compute_config_sdk, mock_compute_config_controller
    ):
        runner = CliRunner()
        result = runner.invoke(create_compute_config, ["missing_config.yaml"])
        _assert_error_message(
            result, message="'missing_config.yaml': No such file or directory",
        )

    def test_config_file_flag_not_found(
        self, fake_compute_config_sdk, mock_compute_config_controller
    ):
        runner = CliRunner()
        result = runner.invoke(create_compute_config, ["-f", "missing_config.yaml"])
        assert isinstance(result.exception, FileNotFoundError)
        assert "No such file or directory: 'missing_config.yaml'" in str(
            result.exception
        )

    def test_config_file_argument(
        self, fake_compute_config_sdk, mock_compute_config_controller
    ):
        runner = CliRunner()
        result = runner.invoke(create_compute_config, [MINIMAL_CONFIG_PATH])
        assert result.exit_code == 0, result.stdout
        assert fake_compute_config_sdk.created_config is None
        mock_compute_config_controller.create.assert_called_once()

    @pytest.mark.parametrize(
        "name", [None, "test-name"],
    )
    @pytest.mark.parametrize(
        "config_file", [MINIMAL_CONFIG_PATH, FULL_CONFIG_PATH],
    )
    def test_config_file_flag(
        self, fake_compute_config_sdk, mock_compute_config_controller, config_file, name
    ):
        runner = CliRunner()
        args = ["-f", config_file]
        if name is not None:
            args += ["-n", name]
        result = runner.invoke(create_compute_config, args)
        assert result.exit_code == 0, result.stdout
        mock_compute_config_controller.create.assert_not_called()
        assert fake_compute_config_sdk.created_config.cloud == "test-cloud-from-file"
        if name is None:
            assert (
                fake_compute_config_sdk.created_name
                == fake_compute_config_sdk.DEFAULT_COMPUTE_CONFIG_NAME
            )
        else:
            assert fake_compute_config_sdk.created_name == name


class TestGet:
    def test_missing_name_and_id(
        self, fake_compute_config_sdk, mock_compute_config_controller
    ):
        runner = CliRunner()
        result = runner.invoke(get_compute_config, [])
        _assert_error_message(
            result,
            message="Either -n/--name or --id/--compute-config-id must be provided",
        )

    def test_name_and_id(self, fake_compute_config_sdk, mock_compute_config_controller):
        runner = CliRunner()
        result = runner.invoke(
            get_compute_config, ["-n", "test-name", "--id", "test-id"]
        )
        _assert_error_message(
            result, message="Only one of name or ID can be provided",
        )

    def test_positional_and_flag_name(
        self, fake_compute_config_sdk, mock_compute_config_controller
    ):
        runner = CliRunner()
        result = runner.invoke(
            get_compute_config, ["-n", "test-name", "test-positional-name"]
        )
        _assert_error_message(
            result, message="Both -n/--name and [COMPUTE_CONFIG_NAME] were provided",
        )

    @pytest.mark.parametrize("include_archived", [False, True])
    @pytest.mark.parametrize("positional_name", [False, True])
    def test_get_by_name(
        self,
        fake_compute_config_sdk,
        mock_compute_config_controller,
        positional_name,
        include_archived,
    ):
        runner = CliRunner()
        if positional_name:
            args = ["test-name"]
        else:
            args = ["-n", "test-name"]

        if include_archived:
            args = ["--include-archived"] + args

        result = runner.invoke(get_compute_config, args)
        assert result.exit_code == 0, result.stdout
        mock_compute_config_controller.get.assert_not_called()
        assert fake_compute_config_sdk.fetched_name == "test-name"
        assert fake_compute_config_sdk.fetched_id is None
        assert fake_compute_config_sdk.fetched_include_archived is include_archived

    @pytest.mark.parametrize("include_archived", [False, True])
    def test_get_by_id(
        self, fake_compute_config_sdk, mock_compute_config_controller, include_archived
    ):
        runner = CliRunner()
        args = ["--id", "test-id"]
        if include_archived:
            args = ["--include-archived"] + args

        result = runner.invoke(get_compute_config, args)
        assert result.exit_code == 0, result.stdout
        mock_compute_config_controller.get.assert_not_called()
        assert fake_compute_config_sdk.fetched_name is None
        assert fake_compute_config_sdk.fetched_id == "test-id"
        assert fake_compute_config_sdk.fetched_include_archived is include_archived

    @pytest.mark.parametrize("include_archived", [False, True])
    @pytest.mark.parametrize("positional_name", [False, True])
    def test_old_format_by_name(
        self,
        fake_compute_config_sdk,
        mock_compute_config_controller,
        positional_name,
        include_archived,
    ):
        runner = CliRunner()
        if positional_name:
            args = ["test-name"]
        else:
            args = ["-n", "test-name"]

        if include_archived:
            args = ["--include-archived"] + args

        args = ["--old-format"] + args

        result = runner.invoke(get_compute_config, args)
        assert result.exit_code == 0, result.stdout
        mock_compute_config_controller.get.assert_called_once_with(
            cluster_compute_name="test-name",
            cluster_compute_id=None,
            include_archived=include_archived,
        )
        assert fake_compute_config_sdk.fetched_name is None
        assert fake_compute_config_sdk.fetched_id is None
        assert fake_compute_config_sdk.fetched_include_archived is None

    @pytest.mark.parametrize("include_archived", [False, True])
    def test_old_format_by_id(
        self, fake_compute_config_sdk, mock_compute_config_controller, include_archived
    ):
        runner = CliRunner()
        args = ["--old-format", "--id", "test-id"]

        if include_archived:
            args = ["--include-archived"] + args

        result = runner.invoke(get_compute_config, args)
        assert result.exit_code == 0, result.stdout
        mock_compute_config_controller.get.assert_called_once_with(
            cluster_compute_name=None,
            cluster_compute_id="test-id",
            include_archived=include_archived,
        )
        assert fake_compute_config_sdk.fetched_name is None
        assert fake_compute_config_sdk.fetched_id is None
        assert fake_compute_config_sdk.fetched_include_archived is None


class TestArchive:
    def test_missing_name_and_id(self, fake_compute_config_sdk):
        runner = CliRunner()
        result = runner.invoke(archive_compute_config, [])
        _assert_error_message(
            result,
            message="Either -n/--name or --id/--compute-config-id must be provided",
        )

    def test_name_and_id(self, fake_compute_config_sdk):
        runner = CliRunner()
        result = runner.invoke(
            archive_compute_config, ["-n", "test-name", "--id", "test-id"]
        )
        _assert_error_message(
            result, message="Only one of name or ID can be provided",
        )

    def test_positional_and_flag_name(self, fake_compute_config_sdk):
        runner = CliRunner()
        result = runner.invoke(
            archive_compute_config, ["-n", "test-name", "test-positional-name"]
        )
        _assert_error_message(
            result, message="Both -n/--name and [COMPUTE_CONFIG_NAME] were provided",
        )

    @pytest.mark.parametrize("positional_name", [False, True])
    def test_get_by_name(self, fake_compute_config_sdk, positional_name):
        runner = CliRunner()
        if positional_name:
            args = ["test-name"]
        else:
            args = ["-n", "test-name"]

        result = runner.invoke(archive_compute_config, args)
        assert result.exit_code == 0, result.stdout
        assert fake_compute_config_sdk.archived_name == "test-name"
        assert fake_compute_config_sdk.archived_id is None

    def test_get_by_id(self, fake_compute_config_sdk):
        runner = CliRunner()
        args = ["--id", "test-id"]

        result = runner.invoke(archive_compute_config, args)
        assert result.exit_code == 0, result.stdout
        assert fake_compute_config_sdk.archived_name is None
        assert fake_compute_config_sdk.archived_id == "test-id"



import pathlib
import subprocess
import tempfile
from typing import Optional, Tuple

import click

from anyscale._private.models.image_uri import ImageURI
from anyscale._private.sdk import _LAZY_SDK_SINGLETONS
from anyscale.cli_logger import BlockLogger
from anyscale.commands.util import convert_kv_strings_to_dict
import anyscale.workspace
from anyscale.workspace._private.workspace_sdk import ANYSCALE_WORKSPACES_SSH_OPTIONS
from anyscale.workspace.commands import _WORKSPACE_SDK_SINGLETON_KEY
from anyscale.workspace.models import WorkspaceConfig, WorkspaceState


log = BlockLogger()  # CLI Logger


def _validate_workspace_name_and_id(
    name: Optional[str], id: Optional[str]  # noqa: A002
):
    if name is None and id is None:
        raise click.ClickException("One of '--name' and '--id' must be provided.")

    if name is not None and id is not None:
        raise click.ClickException("Only one of '--name' and '--id' can be provided.")


@click.group("workspace_v2", help="Anyscale workspace commands V2.")
def workspace_cli() -> None:
    pass


@workspace_cli.command(
    name="create", help="Create a workspace on Anyscale.",
)
@click.option(
    "-f",
    "--config-file",
    required=False,
    default=None,
    type=str,
    help="Path to a YAML config file to deploy. When deploying from a file, import path and arguments cannot be provided. Command-line flags will overwrite values read from the file.",
)
@click.option(
    "-n", "--name", required=False, help="Name of the workspace to create.",
)
@click.option(
    "--image-uri",
    required=False,
    default=None,
    type=str,
    help="Container image to use for the workspace. This is exclusive with --containerfile.",
)
@click.option(
    "--registry-login-secret",
    required=False,
    default=None,
    type=str,
    help="Name or identifier of the secret containing credentials to authenticate to the docker registry hosting the image. "
    "This can only be used when 'image_uri' is specified and the image is not hosted on Anyscale.",
)
@click.option(
    "--containerfile",
    required=False,
    default=None,
    type=str,
    help="Path to a containerfile to build the image to use for the workspace. This is exclusive with --image-uri.",
)
@click.option(
    "--ray-version",
    required=False,
    default=None,
    type=str,
    help="The Ray version (X.Y.Z) to the image specified by --image-uri. This is only used when --image-uri is provided. If not provided, the latest Ray version will be used.",
)
@click.option(
    "--compute-config",
    required=False,
    default=None,
    type=str,
    help="Named compute configuration to use for the workspace.",
)
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workspace. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
)
@click.option(
    "-r",
    "--requirements",
    required=False,
    default=None,
    type=str,
    help="Path to a requirements.txt file containing dependencies for the workspace. These will be installed on top of the image.",
)
@click.option(
    "--env",
    required=False,
    multiple=True,
    type=str,
    help="Environment variables to set for the workspace. The format is 'key=value'. This argument can be specified multiple times. When the same key is also specified in the config file, the value from the command-line flag will overwrite the value from the config file.",
)
def create(  # noqa: PLR0913, PLR0912, C901
    config_file: Optional[str],
    name: Optional[str],
    image_uri: Optional[str],
    registry_login_secret: Optional[str],
    ray_version: Optional[str],
    containerfile: Optional[str],
    compute_config: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    requirements: Optional[str],
    env: Optional[Tuple[str]],
) -> None:
    """Creates a new workspace.

    A name must be provided, either in the file or in the arguments.

    `$ anyscale workspace_v2 create -n my-workspace`

    or add all the information in the config file and do:

    `$ anyscale workspace_v2 create -f config-file.yaml`

    Command-line flags override values in the config file.
    """
    if config_file is not None:
        if not pathlib.Path(config_file).is_file():
            raise click.ClickException(f"Config file '{config_file}' not found.")

        config = WorkspaceConfig.from_yaml(config_file)
    else:
        config = WorkspaceConfig()

    if containerfile and image_uri:
        raise click.ClickException(
            "Only one of '--containerfile' and '--image-uri' can be provided."
        )

    if ray_version and (not image_uri and not containerfile):
        raise click.ClickException(
            "Ray version can only be used with an image or containerfile.",
        )

    if registry_login_secret and (
        not image_uri or ImageURI.from_str(image_uri).is_cluster_env_image()
    ):
        raise click.ClickException(
            "Registry login secret can only be used with an image that is not hosted on Anyscale."
        )

    if name is not None:
        config = config.options(name=name)

    if not config.name:
        raise click.ClickException("Workspace name must be configured")

    if image_uri is not None:
        config = config.options(image_uri=image_uri)

    if registry_login_secret is not None:
        config = config.options(registry_login_secret=registry_login_secret)

    if ray_version is not None:
        config = config.options(ray_version=ray_version)

    if containerfile is not None:
        config = config.options(containerfile=containerfile)

    if compute_config is not None:
        config = config.options(compute_config=compute_config)

    if cloud is not None:
        config = config.options(cloud=cloud)
    if project is not None:
        config = config.options(project=project)

    if requirements is not None:
        if not pathlib.Path(requirements).is_file():
            raise click.ClickException(f"Requirements file '{requirements}' not found.")
        config = config.options(requirements=requirements)
    if env:
        env_dict = convert_kv_strings_to_dict(env)
        if env_dict:
            config = config.options(env_vars=env_dict)

    anyscale.workspace.create(config,)


@workspace_cli.command(name="start", short_help="Starts a workspace.")
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
def start(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Start a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.start(name=name, id=id, cloud=cloud, project=project)


@workspace_cli.command(name="terminate", short_help="Terminate a workspace.")
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
def terminate(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Terminate a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.terminate(name=name, id=id, cloud=cloud, project=project)


@workspace_cli.command(name="status", short_help="Get the status of a workspace.")
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
def status(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """Get the status of a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    status = anyscale.workspace.status(name=name, id=id, cloud=cloud, project=project)
    log.info(status)


@workspace_cli.command(
    name="wait", short_help="Wait for a workspace to reach a certain status."
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.option(
    "--timeout-s",
    required=False,
    default=1800,
    type=float,
    help="The maximum time in seconds to wait for the workspace to reach the desired state. Default to 30 minutes.",
)
@click.option(
    "--state",
    required=False,
    default=WorkspaceState.RUNNING,
    type=str,
    help="The desired terminal state to wait for. Default is 'RUNNING'.",
)
def wait(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    timeout_s: float,
    state: str,
) -> None:
    """Wait for a workspace to reach a terminal state.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    try:
        state = WorkspaceState.validate(state)
    except ValueError as e:
        raise click.ClickException(str(e))
    anyscale.workspace.wait(
        name=name, id=id, cloud=cloud, project=project, timeout_s=timeout_s, state=state
    )


@workspace_cli.command(
    name="ssh",
    short_help="SSH into a workspace.",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
)
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.pass_context
def ssh(
    ctx,
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
) -> None:
    """SSH into a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.

    You may pass extra args for the ssh command, for example to setup port forwarding:
    anyscale workspace_v2 ssh -n workspace-name -- -L 9000:localhost:9000
    """

    _validate_workspace_name_and_id(name=name, id=id)
    assert (
        anyscale.workspace.status(name=name, id=id, cloud=cloud, project=project)
        == WorkspaceState.RUNNING
    ), "Workspace must be running to SSH into it."
    workspace_private_sdk = _LAZY_SDK_SINGLETONS[_WORKSPACE_SDK_SINGLETON_KEY]
    dir_name = workspace_private_sdk.get_default_dir_name(
        name=name, id=id, cloud=cloud, project=project
    )
    command = f"cd {dir_name} && /bin/bash"
    with tempfile.TemporaryDirectory() as tmpdirname:
        host_name, config_file = anyscale.workspace.generate_ssh_config_file(
            name=name, id=id, cloud=cloud, project=project, ssh_config_path=tmpdirname
        )
        args = ctx.args
        ssh_command = (
            ["ssh"]
            + ANYSCALE_WORKSPACES_SSH_OPTIONS
            + [host_name]
            + ["-F", config_file]
            + ["-tt"]
            + (args if args and len(args) > 0 else [""])
            + [f"bash -i -c {command}"]
        )

        subprocess.run(ssh_command, check=False)


@workspace_cli.command(name="run_command", short_help="Run a command in a workspace.")
@click.option(
    "--id", "--workspace-id", required=False, help="Unique ID of the workspace."
)
@click.option("--name", "-n", required=False, help="Name of the workspace.")
@click.option(
    "--cloud",
    required=False,
    default=None,
    type=str,
    help="The Anyscale Cloud to run this workload on. If not provided, the organization default will be used.",
)
@click.option(
    "--project",
    required=False,
    default=None,
    type=str,
    help="Named project to use for the workpsace. If not provided, the default project for the cloud will be used.",
)
@click.argument("command", type=str)
def run_command(
    id: Optional[str],  # noqa: A002
    name: Optional[str],
    cloud: Optional[str],
    project: Optional[str],
    command: str,
) -> None:
    """Run a command in a workspace.

    To specify the workspace by name, use the --name flag. To specify the workspace by id, use the --id flag. Either name or
id should be used, specifying both will result in an error.
    """
    _validate_workspace_name_and_id(name=name, id=id)
    anyscale.workspace.run_command(
        name=name, id=id, cloud=cloud, project=project, command=command
    )



from typing import Optional

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import FineTunedModel
from anyscale.controllers.base_controller import BaseController


class ModelsController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)

        self.log = log
        self.log.open_block("Output")

    def retrieve_model(self, model_id: str) -> FineTunedModel:
        return self.api_client.get_model_api_v2_llm_models_model_id_get(model_id)



import hashlib
import os
import sys
import time
from typing import Any


def fingerprint(
    path: str,
    exclude_dirs: Any = frozenset([]),
    exclude_paths: Any = frozenset([]),
    mtime_hash: bool = True,
) -> str:
    """Generate a fingerprint of a directory.

    Args:
        path (str): The path to fingerprint.
        exclude_dirs (Collection): Set of dir names to exclude.
        exclude_paths (Collection): Set of absolute file paths to exclude.
        mtime_hash (bool): Whether to hash file modification times instead of
            file contents. Fingerprinting is much faster when enabled.
    """
    path = os.path.abspath(os.path.expanduser(path))
    if exclude_paths is None:
        exclude_paths = []
    contents_hasher = hashlib.sha1()

    def add_hash_of_file(fpath: str) -> None:
        if mtime_hash:
            stat = os.stat(fpath)
            contents_hasher.update(str(stat.st_mtime).encode("utf-8"))
        else:
            with open(fpath, "rb") as f:
                for chunk in iter(lambda: f.read(2 ** 20), b""):
                    contents_hasher.update(chunk)

    to_hash = []
    for dirpath, dirs, files in os.walk(path, topdown=True):
        dirs[:] = [d for d in dirs if d not in exclude_dirs]
        to_hash.append((dirpath, sorted(files)))
    for dirpath, filenames in sorted(to_hash):
        has_files = False
        for name in filenames:
            fpath = os.path.join(dirpath, name)
            if fpath not in exclude_paths:
                contents_hasher.update(name.encode("utf-8"))
                add_hash_of_file(fpath)
                has_files = True
        if has_files:
            contents_hasher.update(dirpath.encode("utf-8"))

    return contents_hasher.hexdigest()


if __name__ == "__main__":
    exclude_paths = [os.path.abspath(sys.argv[0])]
    start = time.time()
    print(fingerprint(".", exclude_paths=exclude_paths, mtime_hash=True))
    print("mtime hash", time.time() - start)
    start = time.time()
    print(fingerprint(".", exclude_paths=exclude_paths, mtime_hash=False))
    print("content hash", time.time() - start)



from typing import Optional

from rich.console import Console

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import (
    AttachMachinePoolToCloudRequest,
    CreateMachinePoolRequest,
    CreateMachinePoolResponse,
    DeleteMachinePoolRequest,
    DetachMachinePoolFromCloudRequest,
    ListMachinePoolsResponse,
)
from anyscale.cloud import get_cloud_id_and_name
from anyscale.controllers.base_controller import BaseController


class MachinePoolController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.console = Console()

    def create_machine_pool(self, machine_pool_name: str,) -> CreateMachinePoolResponse:
        response: CreateMachinePoolResponse = self.api_client.create_machine_pool_api_v2_machine_pools_create_post(
            CreateMachinePoolRequest(machine_pool_name=machine_pool_name)
        ).result
        return response

    def delete_machine_pool(
        self, machine_pool_name: str,
    ):
        self.api_client.delete_machine_pool_api_v2_machine_pools_delete_post(
            DeleteMachinePoolRequest(machine_pool_name=machine_pool_name)
        )

    def list_machine_pools(self,) -> ListMachinePoolsResponse:
        response = self.api_client.list_machine_pools_api_v2_machine_pools_get().result
        return response

    def attach_machine_pool_to_cloud(self, machine_pool_name: str, cloud: str):
        cloud_id, _ = get_cloud_id_and_name(self.api_client, cloud_name=cloud)
        self.api_client.attach_machine_pool_to_cloud_api_v2_machine_pools_attach_post(
            AttachMachinePoolToCloudRequest(
                machine_pool_name=machine_pool_name, cloud_id=cloud_id
            )
        )

    def detach_machine_pool_from_cloud(self, machine_pool_name: str, cloud: str):
        cloud_id, _ = get_cloud_id_and_name(self.api_client, cloud_name=cloud)
        self.api_client.detach_machine_pool_from_cloud_api_v2_machine_pools_detach_post(
            DetachMachinePoolFromCloudRequest(
                machine_pool_name=machine_pool_name, cloud_id=cloud_id
            )
        )



import re
from unittest.mock import Mock, patch

from click import ClickException
from google.cloud.certificate_manager_v1.types import (
    Certificate,
    CertificateMap,
    CertificateMapEntry,
)
import pytest
import yaml

from anyscale.cli_logger import BlockLogger, CloudSetupLogger
from anyscale.utils.gcp_managed_setup_utils import (
    append_project_iam_policy,
    configure_firewall_policy,
    create_anyscale_aws_provider,
    create_workload_identity_pool,
    delete_gcp_deployment,
    delete_gcp_tls_certificates,
    delete_workload_identity_pool,
    enable_project_apis,
    GCPDeployment,
    generate_deployment_manager_config,
    get_anyscale_gcp_access_service_acount,
    get_deployment_config,
    get_deployment_resource,
    get_deployment_resources,
    get_or_create_memorystore_gcp,
    get_project_number,
    get_workload_identity_pool,
    remove_firewall_policy_associations,
    update_deployment,
    update_deployment_with_bucket_only,
    wait_for_operation_completion,
)


@pytest.mark.parametrize(
    ("response", "expected_error_message"),
    [
        pytest.param(("200 OK", "project_get.json"), None, id="succeed",),
        pytest.param(
            ("403 Forbidden", None),
            "Error occurred when trying to access the project",
            id="error",
        ),
    ],
)
def test_get_project_number(setup_mock_server, response, expected_error_message):
    factory, tracker = setup_mock_server
    mock_project_id = "anyscale-bridge-deadbeef"
    mock_project_name = "projects/112233445566"
    tracker.reset({".*": [response]})
    if expected_error_message:
        with pytest.raises(ClickException) as e:
            get_project_number(factory, mock_project_id)
        e.match(expected_error_message)
    else:
        assert get_project_number(factory, mock_project_id) == mock_project_name


@pytest.mark.parametrize(
    ("responses", "expected_error_message"),
    [
        pytest.param(
            [
                ("200 OK", "project_get_iam_policy.json"),
                ("200 OK", "project_set_iam_policy.json"),
            ],
            None,
            id="no-role-in-bindings",
        ),
        pytest.param(
            [
                ("200 OK", "project_get_iam_policy_role_exists.json"),
                ("200 OK", "project_set_iam_policy_role_exists.json"),
            ],
            None,
            id="member-not-in-role",
        ),
        pytest.param(
            [
                ("200 OK", "project_set_iam_policy.json"),
                ("200 OK", "project_set_iam_policy.json"),
            ],
            None,
            id="member-already-exists",
        ),
        pytest.param(
            [("400 Bad Request", None)],
            "Failed to set IAM policy for project",
            id="member not found",
        ),
        pytest.param(
            [("403 Forbidden", None)],
            "Failed to set IAM policy for project",
            id="project-not-found",
        ),
    ],
)
def test_append_project_iam_policy(
    setup_mock_server, responses, expected_error_message
):
    factory, tracker = setup_mock_server
    tracker.reset({".*": responses})
    mock_project_id = "mock-project"
    mock_role = "roles/iam.securityAdmin"
    mock_member = "serviceAccount:112233445566@cloudservices.gserviceaccount.com"
    if expected_error_message:
        with pytest.raises(ClickException) as e:
            append_project_iam_policy(factory, mock_project_id, mock_role, mock_member)
        e.match(expected_error_message)
    else:
        updated_policy = append_project_iam_policy(
            factory, mock_project_id, mock_role, mock_member
        )
        bingdings = updated_policy.bindings
        assert mock_role in [binding.role for binding in bingdings]
        assert mock_member in [
            member
            for binding in bingdings
            if binding.role == mock_role
            for member in binding.members
        ]


@pytest.mark.parametrize(
    ("response", "expected_error_message"),
    [
        pytest.param(("200 OK", "enable_api_operation.json"), None, id="succeed",),
        pytest.param(
            ("403 Forbidden", None), "Failed to enable APIs for project", id="error",
        ),
    ],
)
@pytest.mark.parametrize("enable_head_node_fault_tolerance", [True, False])
def test_enable_project_apis(
    setup_mock_server,
    response,
    expected_error_message,
    enable_head_node_fault_tolerance,
):
    factory, tracker = setup_mock_server
    mock_project_id = "mock-project"
    tracker.reset({".*": [response]})
    if expected_error_message:
        with patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ), pytest.raises(ClickException) as e:
            enable_project_apis(
                factory,
                mock_project_id,
                CloudSetupLogger(),
                enable_head_node_fault_tolerance,
            )
        e.match(expected_error_message)
    else:
        with patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):
            assert (
                enable_project_apis(
                    factory,
                    mock_project_id,
                    CloudSetupLogger(),
                    enable_head_node_fault_tolerance,
                )
                is None
            )


@pytest.mark.parametrize(
    ("response", "expected_error_message", "pool_exists"),
    [
        pytest.param(
            ("200 OK", "get_workload_identity_pool.json"), None, True, id="succeed",
        ),
        pytest.param(("404 Not Found", None), None, False, id="NotFound"),
        pytest.param(
            ("403 Forbidden", None),
            "Failed to get Workload Identity Provider Pool.",
            True,
            id="error",
        ),
    ],
)
def test_get_workload_identity_pool(
    setup_mock_server, response, pool_exists, expected_error_message
):
    factory, tracker = setup_mock_server
    mock_project_id = "anyscale-bridge-deadbeef"
    mock_pool_id = "mock-pool-id"
    tracker.reset({".*": [response]})
    if expected_error_message:
        with pytest.raises(ClickException) as e:
            get_workload_identity_pool(factory, mock_project_id, mock_pool_id)
        e.match(expected_error_message)
    elif pool_exists:
        assert (
            get_workload_identity_pool(factory, mock_project_id, mock_pool_id)
            == mock_pool_id
        )
    else:
        assert (
            get_workload_identity_pool(factory, mock_project_id, mock_pool_id) is None
        )


@pytest.mark.parametrize(
    ("response", "expected_error_message", "service_account_exists"),
    [
        pytest.param(("200 OK", "get_service_account.json"), None, True, id="succeed",),
        pytest.param(("404 Not Found", None), None, False, id="NotFound"),
        pytest.param(
            ("403 Forbidden", None), "Failed to get service account: ", True, id="error"
        ),
    ],
)
def test_get_anyscale_gcp_access_service_acount(
    setup_mock_server, response, service_account_exists, expected_error_message
):
    factory, tracker = setup_mock_server
    mock_service_account = (
        "anyscale-access@anyscale-bridge-deadbeef.iam.gserviceaccount.com"
    )
    tracker.reset({".*": [response]})
    if expected_error_message:
        with pytest.raises(ClickException) as e:
            get_anyscale_gcp_access_service_acount(factory, mock_service_account)
        e.match(expected_error_message)
    elif service_account_exists:
        assert (
            get_anyscale_gcp_access_service_acount(factory, mock_service_account)
            == mock_service_account
        )
    else:
        assert (
            get_anyscale_gcp_access_service_acount(factory, mock_service_account)
            is None
        )


@pytest.mark.parametrize(
    ("response", "expected_log_message"),
    [
        pytest.param(
            ("200 OK", "create_workload_identity_pool.json"), None, id="succeed",
        ),
        pytest.param(("409 conflict", None), "already exists", id="pool-exists",),
        pytest.param(("418 I'm a teapot", None), "Error occurred", id="error",),
    ],
)
def test_create_workload_identity_pool(
    setup_mock_server, response, expected_log_message, capsys
):
    factory, tracker = setup_mock_server
    mock_project_id = "anyscale-bridge-deadbeef"
    mock_project_number = "123456789"
    mock_pool_id = "mock-pool-id"
    display_name = "mock pool"
    description = "mock provider pool"
    tracker.reset({".*": [response]})
    if expected_log_message:
        with pytest.raises(ClickException) as e, patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):
            create_workload_identity_pool(
                factory,
                mock_project_id,
                mock_pool_id,
                BlockLogger(),
                display_name,
                description,
            )
        e.match("Failed to create Workload Identity Provider Pool")
        _, err = capsys.readouterr()
        assert expected_log_message in err

    else:
        with patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):
            assert (
                create_workload_identity_pool(
                    factory,
                    mock_project_id,
                    mock_pool_id,
                    BlockLogger(),
                    display_name,
                    description,
                )
                == f"projects/{mock_project_number}/locations/global/workloadIdentityPools/{mock_pool_id}"
            )


@pytest.mark.parametrize(
    ("response", "expected_error_message"),
    [
        pytest.param(
            ("200 OK", "create_workload_identity_provider.json"), None, id="succeed",
        ),
        pytest.param(("409 conflict", None), "already exists", id="pool-exists",),
        pytest.param(("404 Not Found", None), "Error occurred", id="error",),
    ],
)
def test_create_anyscale_aws_provider(
    setup_mock_server, response, expected_error_message, capsys
):
    factory, tracker = setup_mock_server
    mock_project_number = "123456789"
    mock_pool_id = f"projects/{mock_project_number}/locations/global/workloadIdentityPools/mock-pool-id"
    mock_provider_id = "mock-provider"
    mock_aws_account = "123456"
    mock_display_name = "mock provider"
    mock_org_id = "mock_org_id"
    tracker.reset({".*": [response]})
    if expected_error_message:
        with pytest.raises(ClickException) as e, patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):
            create_anyscale_aws_provider(
                factory,
                mock_org_id,
                mock_pool_id,
                mock_provider_id,
                mock_aws_account,
                mock_display_name,
                BlockLogger(),
            )
        e.match("Failed to create Anyscale AWS Workload Identity Provider")
        _, err = capsys.readouterr()
        assert expected_error_message in err
    else:
        with patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):
            assert (
                create_anyscale_aws_provider(
                    factory,
                    mock_org_id,
                    mock_pool_id,
                    mock_provider_id,
                    mock_aws_account,
                    mock_display_name,
                    BlockLogger(),
                )
                == f"{mock_pool_id}/providers/{mock_provider_id}"
            )


@pytest.mark.parametrize(
    ("response", "expected_log_message", "deletion_succeed"),
    [
        pytest.param(
            ("200 OK", "delete_workload_identity_pool.json"),
            "Deleted workload identity pool",
            True,
            id="succeed",
        ),
        pytest.param(
            ("403 Forbidden", None),
            "Error occurred when trying to delete workload identity pool",
            False,
            id="error1",
        ),
        pytest.param(("404 Not Found", None), None, False, id="error2",),
    ],
)
def test_delete_workload_identity_pool(
    setup_mock_server, response, deletion_succeed, expected_log_message, capsys
):
    factory, tracker = setup_mock_server
    mock_project_number = "123456789"
    mock_pool_id = "mock-pool-id"
    mock_pool_name = f"projects/{mock_project_number}/locations/global/workloadIdentityPools/{mock_pool_id}"
    tracker.reset({".*": [response]})
    if deletion_succeed:
        with patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):
            delete_workload_identity_pool(
                factory, mock_pool_name, BlockLogger(),
            )
            _, log = capsys.readouterr()
            assert expected_log_message in log
    elif expected_log_message:
        with pytest.raises(ClickException) as e, patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):
            delete_workload_identity_pool(
                factory, mock_pool_name, BlockLogger(),
            )
        e.match(expected_log_message)
    else:
        # not found
        with patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):
            assert (
                delete_workload_identity_pool(factory, mock_pool_name, BlockLogger(),)
                is None
            )


@pytest.mark.parametrize("service_type", ["workload_identity_pool", "provider"])
@pytest.mark.parametrize(
    ("response", "expected_error_message"),
    [
        pytest.param(
            ("200 OK", "create_workload_identity_pool.json"),
            "did not complete within the timeout period",
            id="timeout",
        ),
        pytest.param(("200 OK", "operation_completed.json"), None, id="succeed",),
        pytest.param(
            ("200 OK", "operation_error.json"), "encountered an error", id="error",
        ),
    ],
)
def test_wait_for_operation_completion(
    setup_mock_server, response, expected_error_message, service_type
):
    factory, tracker = setup_mock_server
    mock_project_name = "projects/112233445566"
    mock_pool_id = "mock-pool-id"
    mock_provider_id = "mock-provider"
    if service_type == "workload_identity_pool":
        service = (
            factory.build("iam", "v1").projects().locations().workloadIdentityPools()
        )
        mock_operation_id = f"{mock_project_name}/locations/global/workloadIdentityPools/{mock_pool_id}/operations/mock_operation_id"
    elif service_type == "provider":
        service = (
            factory.build("iam", "v1")
            .projects()
            .locations()
            .workloadIdentityPools()
            .providers()
        )
        mock_operation_id = f"{mock_project_name}/locations/global/workloadIdentityPools/{mock_pool_id}/providers/{mock_provider_id}/operations/mock_operation_id"
    tracker.reset({".*": [response]})
    if expected_error_message:
        with patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils.time",
            time=Mock(side_effect=list(range(10))),  # only iterates once
            sleep=Mock(),
        ), pytest.raises(ClickException) as e:
            wait_for_operation_completion(
                service,
                {"name": mock_operation_id},
                "test",
                timeout=2,
                polling_interval=10,
            )
        e.match(expected_error_message)
    else:
        with patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils.time",
            time=Mock(side_effect=list(range(10))),
            sleep=Mock(),
        ):
            assert (
                wait_for_operation_completion(
                    service, {"name": mock_operation_id}, "test",
                )
                is None
            )


@pytest.mark.parametrize(
    ("responses", "expected_error_message"),
    [
        pytest.param(
            [
                ("200 OK", "add_firewall_policy_rule.json"),
                ("200 OK", "add_firewall_policy_rule.json"),
                ("200 OK", "associate_firewall_policy.json"),
                ("200 OK", "add_firewall_policy_rule_done.json"),
                ("200 OK", "add_firewall_policy_rule_done.json"),
                ("200 OK", "associate_firewall_policy_done.json"),
            ],
            None,
            id="succeed",
        ),
        pytest.param(
            [
                ("200 OK", "add_firewall_policy_rule.json"),
                ("200 OK", "add_firewall_policy_rule.json"),
                ("200 OK", "associate_firewall_policy.json"),
                ("200 OK", "add_firewall_policy_rule.json"),
            ],
            "Timeout when trying to configure firewall policy",
            id="timeout",
        ),
        pytest.param(
            [
                ("200 OK", "add_firewall_policy_rule.json"),
                ("200 OK", "add_firewall_policy_rule.json"),
                ("200 OK", "associate_firewall_policy.json"),
                ("200 OK", "add_firewall_policy_rule_error.json"),
            ],
            "Failed to configure firewall policy ",
            id="error",
        ),
        pytest.param(
            [("400 Bad Request", None)],
            "Failed to configure firewall policy",
            id="bad-request",
        ),
        pytest.param(
            [
                ("200 OK", "add_firewall_policy_rule.json"),
                ("200 OK", "add_firewall_policy_rule.json"),
                ("404 Not Found", None),
            ],
            "Failed to configure firewall policy",
            id="not-found",
        ),
    ],
)
def test_configure_firewall_policy(
    setup_mock_server, responses, expected_error_message
):
    factory, tracker = setup_mock_server
    mock_project_id = "mock-project"
    mock_vpc_name = "mock-vpc"
    mock_firewall_policy = "mock-fp"
    mock_subnet_cidr = "10.0.0.0/20"
    tracker.reset({".*": responses})
    if expected_error_message:
        with pytest.raises(ClickException) as e:
            configure_firewall_policy(
                factory,
                mock_vpc_name,
                mock_project_id,
                mock_firewall_policy,
                mock_subnet_cidr,
            )
        e.match(expected_error_message)
    else:
        configure_firewall_policy(
            factory,
            mock_vpc_name,
            mock_project_id,
            mock_firewall_policy,
            mock_subnet_cidr,
        )


@pytest.mark.parametrize(
    ("responses", "expected_error_message"),
    [
        pytest.param(
            [("404 Not Found", None)], "Failed to get deployment", id="failed",
        ),
        pytest.param(
            [("200 OK", "deployment_get.json"), ("200 OK", "manifest_get.json")],
            None,
            id="succeed",
        ),
    ],
)
def test_get_deployment_resources(setup_mock_server, responses, expected_error_message):
    factory, tracker = setup_mock_server
    mock_deployment_name = "mock-deployment"
    mock_project_id = "mock-project"
    mock_anyscale_access_service_account_name = "anyscale-access-cld-congtest"
    tracker.reset({".*": responses})
    if expected_error_message:
        with pytest.raises(ClickException) as e:
            get_deployment_resources(
                factory,
                mock_deployment_name,
                mock_project_id,
                mock_anyscale_access_service_account_name,
            )
        e.match(expected_error_message)
    else:
        assert get_deployment_resources(
            factory,
            mock_deployment_name,
            mock_project_id,
            mock_anyscale_access_service_account_name,
        ) == {
            "compute.v1.network": "vpc-cld-congtest",
            "compute.v1.subnetwork": "subnet-cld-congtest",
            "gcp-types/compute-v1:networkFirewallPolicies": "firewall-policy-cld-congtest",
            "filestore_instance": "filestore-cld-congtest",
            "filestore_location": "us-west1",
            "gcp-types/file-v1beta1:projects.locations.instances": "filestore-cld-congtest",
            "storage.v1.bucket": "storage-bucket-cld-congtest",
            "iam.v1.serviceAccount": "instance-cld-congtest",
        }


@pytest.mark.parametrize(
    ("responses", "expected_error_message"),
    [
        pytest.param(
            [("200 OK", "deployment_get.json"), ("200 OK", "deployment_delete.json")],
            None,
            id="no-bucket-delete-succeed",
        ),
        pytest.param(
            [("200 OK", "deployment_get.json"), ("403 Forbidden", None)],
            "Failed to delete deployment",
            id="no-bucket-delete-fail",
        ),
        pytest.param([("404 Not Found", None)], None, id="no-deployment"),
        pytest.param(
            [("403 Forbidden", None)], "Failed to get deployment", id="forbidden"
        ),
    ],
)
def test_delete_gcp_deployment(setup_mock_server, responses, expected_error_message):
    factory, tracker = setup_mock_server
    mock_project_id = "mock-project"
    mock_deployment = "mock-deployment"
    tracker.reset({".*": responses})

    if expected_error_message:

        with pytest.raises(ClickException) as e, patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):
            delete_gcp_deployment(factory, mock_project_id, mock_deployment)
        e.match(expected_error_message)
    else:
        with patch.multiple(
            "anyscale.utils.gcp_managed_setup_utils",
            wait_for_operation_completion=Mock(),
        ):

            assert (
                delete_gcp_deployment(factory, mock_project_id, mock_deployment) is None
            )


def test_update_deployment_with_bucket_only():
    factory = Mock()
    mock_project_id = "mock-project"
    mock_deployment = "mock-deployment"
    mock_fingerprint = "mock"
    mock_config_content = '\nresources:\n- name: vpc-cld-congtest\n  type: compute.v1.network\n  properties:\n    name: vpc-cld-congtest\n    autoCreateSubnetworks: False\n- name: subnet-cld-congtest\n  type: compute.v1.subnetwork\n  properties:\n    name: subnet-cld-congtest\n    ipCidrRange: 10.0.0.0/20\n    region: us-west1\n    network: $(ref.vpc-cld-congtest.selfLink)\n- name: firewall-policy-cld-congtest\n  type: gcp-types/compute-v1:networkFirewallPolicies\n  properties:\n    parentId: projects/mock-project/regions/us-west1\n    name: firewall-policy-cld-congtest\n    description: "firewall policy for Anyscale managed cloud cld-congtest"\n- name: filestore-cld-congtest\n  type: gcp-types/file-v1beta1:projects.locations.instances\n  properties:\n    instanceId: filestore-cld-congtest\n    parent: projects/mock-project/locations/us-west1\n    tier: ENTERPRISE\n    networks:\n      - network: projects/mock-project/global/networks/$(ref.vpc-cld-congtest.name)\n    fileShares:\n      - name: anyscale_fileshare\n        capacityGb: 1024\n- name: storage-bucket-cld-congtest\n  type: storage.v1.bucket\n  properties:\n    location: us-west1\n    storageClass: REGIONAL\n    iamConfiguration:\n      uniformBucketLevelAccess:\n        enabled: True\n      publicAccessPrevention: enforced\n    cors:\n    - origin: ["https://console.anyscale-staging.com"]\n      responseHeader: ["*"]\n      method: ["GET"]\n- name: anyscale-access-cld-congtest\n  type: iam.v1.serviceAccount\n  properties:\n    accountId: anyscale-access-cld-congtest\n    displayName: test anyscale account for managed setup\n    projectId: mock-project\n  accessControl:\n    gcpIamPolicy:\n        bindings:\n        - role: roles/iam.serviceAccountTokenCreator\n          members:\n          - serviceAccount:anyscale-access-cld-congtest@mock-project.iam.gserviceaccount.com \n- name: iam-policy-anyscale-access-project-cld-congtest\n  type: gcp-types/cloudresourcemanager-v1:virtual.projects.iamMemberBinding\n  metadata:\n    dependsOn:\n      - anyscale-access-cld-congtest\n  properties: \n    resource: mock-project\n    role: roles/editor\n    member: "serviceAccount:$(ref.anyscale-access-cld-congtest.email)"\n- name: instance-cld-congtest\n  type: iam.v1.serviceAccount\n  properties:\n    accountId: instance-cld-congtest\n    displayName: test instance account for managed setup\n    projectId: mock-project\n- name: iam-policy-instance-project-cld-congtest\n  type: gcp-types/cloudresourcemanager-v1:virtual.projects.iamMemberBinding\n  metadata:\n    dependsOn:\n      - instance-cld-congtest\n  properties: \n    resource: mock-project\n    role: roles/artifactregistry.reader\n    member: "serviceAccount:$(ref.instance-cld-congtest.email)"\n- name: bucket-iam-anyscale-acess-cld-congtest\n  type: gcp-types/storage-v1:virtual.buckets.iamMemberBinding\n  properties:\n    bucket: $(ref.storage-bucket-cld-congtest.name)\n    member: serviceAccount:$(ref.anyscale-access-cld-congtest.email)\n    role: roles/storage.admin\n  metadata:\n    dependsOn:\n      - anyscale-access-cld-congtest\n- name: bucket-iam-instance-cld-congtest\n  type: gcp-types/storage-v1:virtual.buckets.iamMemberBinding\n  properties:\n    bucket: $(ref.storage-bucket-cld-congtest.name)\n    member: serviceAccount:$(ref.instance-cld-congtest.email)\n    role: roles/storage.admin\n  metadata:\n    dependsOn:\n      - instance-cld-congtest\n    '
    mock_gcp_deployment = GCPDeployment(
        deployment_name=mock_deployment,
        fingerprint=mock_fingerprint,
        config_content=mock_config_content,
    )

    mock_get_deployment_config = Mock(return_value=mock_gcp_deployment)
    mock_update_deployment = Mock()

    with patch.multiple(
        "anyscale.utils.gcp_managed_setup_utils",
        wait_for_operation_completion=Mock(),
        get_deployment_config=mock_get_deployment_config,
        update_deployment=mock_update_deployment,
    ):

        assert (
            update_deployment_with_bucket_only(
                factory, mock_project_id, mock_deployment
            )
            is None
        )
        mock_get_deployment_config.assert_called_once_with(
            factory, mock_project_id, mock_deployment
        )
        expected_config_content = "resources:\n- name: storage-bucket-cld-congtest\n  properties:\n    cors:\n    - method:\n      - GET\n      origin:\n      - https://console.anyscale-staging.com\n      responseHeader:\n      - '*'\n    iamConfiguration:\n      publicAccessPrevention: enforced\n      uniformBucketLevelAccess:\n        enabled: true\n    location: us-west1\n    storageClass: REGIONAL\n  type: storage.v1.bucket\n"
        mock_update_deployment.assert_called_once_with(
            factory,
            mock_project_id,
            mock_deployment,
            mock_fingerprint,
            expected_config_content,
        )


@pytest.mark.parametrize(
    ("responses", "expected_error_message"),
    [
        pytest.param([("404 Not Found", None)], None, id="no-firewall",),
        pytest.param(
            [
                ("200 OK", "firewall_policy_get.json"),
                ("200 OK", "remove_firewall_vpc_association.json"),
                ("200 OK", "remove_firewall_vpc_association_done.json"),
            ],
            None,
            id="succeed",
        ),
        pytest.param(
            [("200 OK", "firewall_policy_get.json"), ("400 Bad Request", None)],
            None,
            id="no-association",
        ),
        pytest.param(
            [("200 OK", "firewall_policy_get.json"), ("403 Forbidden", None)],
            "Failed to remove firewall policy association.",
            id="forbidden",
        ),
        pytest.param(
            [
                ("200 OK", "firewall_policy_get.json"),
                ("200 OK", "remove_firewall_vpc_association.json"),
                ("200 OK", "remove_firewall_vpc_association.json"),
            ],
            "Timeout",
            id="timeout",
        ),
        pytest.param(
            [
                ("200 OK", "firewall_policy_get.json"),
                ("200 OK", "remove_firewall_vpc_association.json"),
                ("200 OK", "remove_firewall_vpc_association_error.json"),
            ],
            "Failed to remove",
            id="error",
        ),
    ],
)
def test_remove_firewall_policy_associations(
    setup_mock_server, responses, expected_error_message
):
    factory, tracker = setup_mock_server
    mock_project_id = "mock-project"
    mock_firewall_policy = "mock-firewall-policy"
    tracker.reset({".*": responses})
    if expected_error_message:
        with pytest.raises(ClickException) as e:
            remove_firewall_policy_associations(
                factory, mock_project_id, mock_firewall_policy
            )
        e.match(expected_error_message)
    else:
        assert (
            remove_firewall_policy_associations(
                factory, mock_project_id, mock_firewall_policy
            )
            is None
        )


def test_delete_gcp_tls_certificates():
    mock_factory = Mock()
    mock_certificate_manager_client = Mock()
    mock_operation_client = Mock()
    project_id = "project_id"
    cloud_id = "cloud_id"
    location_path = "projects/project_id/locations/global"
    certificate_id = "certificate_id"
    certificate_map_entry_path = (
        "projects/project_id/locations/global/certificateMaps/certificate_map_entry_id"
    )
    certificate_map_entry_name = f"projects/project_id/locations/global/certificateMaps/{certificate_id}/certificateMapEntries/certificate_map_entry_id"
    certificate_name = (
        f"projects/project_id/locations/global/certificates/{certificate_id}"
    )
    certificate_map_name = (
        f"projects/project_id/locations/global/certificateMaps/{certificate_id}"
    )

    anyscale_cloud_label = {}
    anyscale_cloud_label["anyscale-cloud-id"] = cloud_id

    # Mock certificate_manager_client.common_location_path
    mock_certificate_manager_client.common_location_path = Mock(
        return_value=location_path
    )

    # Mock certificate_manager_client.list_certificates
    certificate = Certificate(name=certificate_name, labels=anyscale_cloud_label)
    mock_certificate_manager_client.list_certificates = Mock(return_value=[certificate])

    # Mock certificate_manager_client.list_certificate_maps
    certificate_map = CertificateMap(
        name=certificate_map_name, labels=anyscale_cloud_label
    )
    mock_certificate_manager_client.list_certificate_maps = Mock(
        return_value=[certificate_map]
    )

    # Mock certificate_manager_client.certificate_map_path
    mock_certificate_manager_client.certificate_map_path = Mock(
        return_value=certificate_map_entry_path
    )

    # Mock certificate_manager_client.list_certificate_map_entries
    certificate_map_entry = CertificateMapEntry(
        name=certificate_map_entry_name, labels=anyscale_cloud_label
    )
    mock_certificate_manager_client.list_certificate_map_entries = Mock(
        return_value=[certificate_map_entry]
    )

    # Mock delete resource requests
    mock_certificate_manager_client.delete_certificate_map_entry = Mock()
    mock_certificate_manager_client.delete_certificate_map = Mock()
    mock_certificate_manager_client.delete_certificate = Mock()

    # Create Mock clients
    mock_factory.compute_v1.GlobalOperationsClient = Mock(
        return_value=mock_operation_client
    )
    mock_factory.certificate_manager_v1.CertificateManagerClient = Mock(
        return_value=mock_certificate_manager_client
    )

    delete_gcp_tls_certificates(mock_factory, project_id, cloud_id)

    # Assert function calls
    mock_certificate_manager_client.common_location_path.assert_called_once_with(
        project_id, "global"
    )
    mock_certificate_manager_client.list_certificates.assert_called_once_with(
        parent=location_path
    )
    mock_certificate_manager_client.list_certificate_maps.assert_called_once_with(
        parent=location_path
    )
    mock_certificate_manager_client.certificate_map_path.assert_called_once_with(
        project_id, "global", certificate_id
    )
    mock_certificate_manager_client.list_certificate_map_entries.assert_called_once_with(
        parent=certificate_map_entry_path
    )
    mock_certificate_manager_client.delete_certificate_map_entry.assert_called_once_with(
        name=certificate_map_entry_name
    )
    mock_certificate_manager_client.delete_certificate_map.assert_called_once_with(
        name=certificate_map_name
    )
    mock_certificate_manager_client.delete_certificate.assert_called_once_with(
        name=certificate_name
    )


@pytest.mark.parametrize("enable_head_node_fault_tolerance", [True, False])
def test_generate_deployment_manager_config(enable_head_node_fault_tolerance):
    mock_region = "mock-region"
    mock_project_id = "mock-project"
    mock_cloud_id = "cld_mock"
    mock_service_account_name = "anyscale-access-mock"
    workload_identity_pool_name = (
        "projects/112233445566/locations/global/workloadIdentityPools/mock-pool"
    )
    anyscale_aws_account = "123456"
    organization_id = "org_mock"

    generated_config = generate_deployment_manager_config(
        mock_region,
        mock_project_id,
        mock_cloud_id,
        mock_service_account_name,
        workload_identity_pool_name,
        anyscale_aws_account,
        organization_id,
        enable_head_node_fault_tolerance,
    )

    # verify it's a valid yaml file
    yaml.safe_load(generated_config)

    # verify all ${val} are substituted
    assert re.search(r"\$\{.*\}", generated_config) is None

    if enable_head_node_fault_tolerance:
        assert "gcp-types/redis-v1:projects.locations.instances" in generated_config
    else:
        assert "gcp-types/redis-v1:projects.locations.instances" not in generated_config


@pytest.mark.parametrize(
    ("redis_instance_exists", "expected_error"),
    [
        pytest.param(True, False, id="redis-instance-exists"),
        pytest.param(False, True, id="error"),
        pytest.param(False, False, id="happy-path"),
    ],
)
@pytest.mark.parametrize("yes", [True, False])
def test_get_or_create_memorystore_gcp(redis_instance_exists, yes, expected_error):
    factory = Mock()
    mock_cloud_id = "mock-cloud-id"
    mock_deployment = mock_cloud_id
    mock_project_id = "mock-project"
    mock_resource_name = f"redis-{mock_deployment}"
    mock_region = "mock-region"

    mock_get_deployment_config = Mock(
        return_value=GCPDeployment(
            deployment_name=mock_deployment, fingerprint="mock", config_content="mock",
        )
    )
    mock_update_deployment = Mock()
    mock_get_deployment_resource = Mock()
    if redis_instance_exists:
        mock_get_deployment_resource.return_value = {"name": mock_resource_name}
    elif expected_error:
        mock_get_deployment_resource.side_effect = [None, None]
    else:
        mock_get_deployment_resource.side_effect = [None, {"name": mock_resource_name}]

    with patch.multiple(
        "anyscale.utils.gcp_managed_setup_utils",
        get_deployment_resource=mock_get_deployment_resource,
        get_deployment_config=mock_get_deployment_config,
        update_deployment=mock_update_deployment,
        confirm=Mock(),
    ):
        if expected_error:
            with pytest.raises(ClickException) as e:
                get_or_create_memorystore_gcp(
                    factory,
                    mock_cloud_id,
                    mock_deployment,
                    mock_project_id,
                    mock_region,
                    CloudSetupLogger(),
                    yes,
                )
            e.match("Failed to create Memorystore instance in deployment")
        else:
            assert (
                get_or_create_memorystore_gcp(
                    factory,
                    mock_cloud_id,
                    mock_deployment,
                    mock_project_id,
                    mock_region,
                    CloudSetupLogger(),
                    yes,
                )
                == f"projects/{mock_project_id}/locations/{mock_region}/instances/{mock_resource_name}"
            )
            if redis_instance_exists:
                mock_get_deployment_resource.assert_called_once()
                mock_get_deployment_config.assert_not_called()
                mock_update_deployment.assert_not_called()
            else:
                assert mock_get_deployment_resource.call_count == 2
                assert mock_get_deployment_config.call_count == 2
                if yes:
                    mock_update_deployment.assert_called_once()
                else:
                    assert mock_get_deployment_resource.call_count == 2


@pytest.mark.parametrize(
    ("response", "expected_error", "expected_none"),
    [
        pytest.param(("404 Not Found", None), False, True, id="404-not-found",),
        pytest.param(("403 Forbidden", None), True, True, id="403-forbidden",),
        pytest.param(
            ("200 OK", "deployment_manager_resources_get.json"),
            False,
            False,
            id="happy-path",
        ),
    ],
)
def test_get_deployment_resource(
    setup_mock_server, response, expected_error, expected_none
):
    factory, tracker = setup_mock_server
    mock_project_id = "mock-project"
    mock_deployment_name = "mock-deployment"
    mock_resource_name = f"redis-{mock_deployment_name}"
    tracker.reset({".*": [response]})
    if expected_error:
        with pytest.raises(ClickException) as e:
            get_deployment_resource(
                factory, mock_project_id, mock_deployment_name, mock_resource_name
            )
        e.match("Failed to get resource")
    else:
        assert (
            get_deployment_resource(
                factory, mock_project_id, mock_deployment_name, mock_resource_name
            )
            is None
        ) == expected_none


@pytest.mark.parametrize(
    ("responses", "expected_error"),
    [
        pytest.param(
            [("200 OK", "deployment_get.json"), ("200 OK", "manifest_get.json")],
            False,
            id="happy-path",
        ),
        pytest.param([("404 Not Found", None)], True, id="no-deployment",),
        pytest.param([("403 Forbidden", None)], True, id="forbidden",),
        pytest.param(
            [("200 OK", "deployment_get.json"), ("403 Forbidden", None)],
            True,
            id="manifest-failed",
        ),
    ],
)
def test_get_deployment_config(setup_mock_server, responses, expected_error):
    factory, tracker = setup_mock_server
    mock_project_id = "mock-project"
    mock_deployment_name = "mock-deployment"
    tracker.reset({".*": responses})
    if expected_error:
        with pytest.raises(ClickException) as e:
            get_deployment_config(factory, mock_project_id, mock_deployment_name)
        e.match("Failed to get deployment config from deployment")
    else:
        assert (
            get_deployment_config(factory, mock_project_id, mock_deployment_name)
            is not None
        )


@pytest.mark.parametrize(
    ("responses", "expected_error"),
    [
        pytest.param([("200 OK", "deployment_update.json")], False, id="happy-path",),
        pytest.param([("404 Not Found", None)], True, id="404",),
        pytest.param([("403 Forbidden", None)], True, id="403",),
    ],
)
def test_update_deployment(setup_mock_server, responses, expected_error):
    factory, tracker = setup_mock_server
    mock_project_id = "mock-project"
    mock_deployment_name = "mock-deployment"
    mock_fingerprint = "mock"
    mock_config_content = "mock"
    tracker.reset({".*": responses})
    with patch.multiple(
        "anyscale.utils.gcp_managed_setup_utils", wait_for_operation_completion=Mock(),
    ):
        if expected_error:
            with pytest.raises(ClickException) as e:
                update_deployment(
                    factory,
                    mock_project_id,
                    mock_deployment_name,
                    mock_fingerprint,
                    mock_config_content,
                )
            e.match("Failed")
        else:
            update_deployment(
                factory,
                mock_project_id,
                mock_deployment_name,
                mock_fingerprint,
                mock_config_content,
            )



import json
import logging
import os
from typing import Any

import click

from anyscale.cli_logger import BlockLogger
from anyscale.commands.anyscale_api.api_commands import anyscale_api
from anyscale.commands.auth_commands import auth_cli
from anyscale.commands.cloud_commands import cloud_cli
from anyscale.commands.cluster_commands import cluster_cli
from anyscale.commands.cluster_env_commands import cluster_env_cli
from anyscale.commands.compute_config_commands import compute_config_cli
from anyscale.commands.config_commands import config_cli
from anyscale.commands.exec_commands import anyscale_exec
from anyscale.commands.experimental_integrations_commands import (
    experimental_integrations_cli,
)
from anyscale.commands.fine_tune_commands import fine_tune_cli
from anyscale.commands.image_commands import image_cli
from anyscale.commands.job_commands import job_cli
from anyscale.commands.list_commands import list_cli
from anyscale.commands.llm.group import llm_cli
from anyscale.commands.login_commands import anyscale_login, anyscale_logout
from anyscale.commands.logs_commands import log_cli
from anyscale.commands.machine_commands import machine_cli
from anyscale.commands.machine_pool_commands import machine_pool_cli
from anyscale.commands.migrate_commands import migrate_cli
from anyscale.commands.project_commands import anyscale_init, project_cli
from anyscale.commands.schedule_commands import schedule_cli
from anyscale.commands.service_account_commands import service_account_cli
from anyscale.commands.service_commands import service_cli
from anyscale.commands.session_commands_hidden import session_cli
from anyscale.commands.workspace_commands import workspace_cli
from anyscale.commands.workspace_commands_v2 import workspace_cli as workspace_cli_v2
import anyscale.conf
from anyscale.utils.cli_version_check_util import log_warning_if_version_needs_upgrade


logger = logging.getLogger(__name__)
logging.getLogger("botocore").setLevel(logging.CRITICAL)

log = BlockLogger()  # CLI Logger

if anyscale.conf.AWS_PROFILE is not None:
    logger.info("Using AWS profile %s", anyscale.conf.AWS_PROFILE)
    os.environ["AWS_PROFILE"] = anyscale.conf.AWS_PROFILE


class AliasedGroup(click.Group):
    # This is from https://stackoverflow.com/questions/46641928/python-click-multiple-command-names
    def get_command(self, ctx: Any, cmd_name: str) -> Any:
        if cmd_name in ALIASES:
            cmd_name = ALIASES[cmd_name].name
        return super().get_command(ctx, cmd_name)


@click.group(
    "anyscale",
    invoke_without_command=True,
    no_args_is_help=True,
    cls=AliasedGroup,
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.option(
    "--version",
    "-v",
    "version_flag",
    is_flag=True,
    default=False,
    help="Current anyscale version.",
)
@click.option(
    "--json",
    "show_json",
    is_flag=True,
    default=False,
    help="Return output as json, for use with --version.",
)
@click.pass_context
def cli(ctx: Any, version_flag: bool, show_json: bool) -> None:
    if version_flag:
        ctx.invoke(version_cli, show_json=show_json)
    log_warning_if_version_needs_upgrade()


@click.command(name="version", help="Display version of the anyscale CLI.")
@click.option(
    "--json", "show_json", is_flag=True, default=False, help="Return output as json."
)
def version_cli(show_json: bool) -> None:
    if show_json:
        print(json.dumps({"version": anyscale.__version__}))
    else:
        print(anyscale.__version__)


@cli.command(
    name="help", help="Display help documentation for anyscale CLI.", hidden=True
)
@click.pass_context
def anyscale_help(ctx: Any) -> None:
    print(ctx.parent.get_help())


cli.add_command(session_cli)
cli.add_command(cloud_cli)
cli.add_command(config_cli)
cli.add_command(migrate_cli)
cli.add_command(project_cli)
cli.add_command(version_cli)
cli.add_command(list_cli)
cli.add_command(cluster_env_cli)
cli.add_command(job_cli)
cli.add_command(schedule_cli)
cli.add_command(service_cli)
cli.add_command(cluster_cli)
cli.add_command(workspace_cli)
cli.add_command(workspace_cli_v2)
cli.add_command(experimental_integrations_cli)
cli.add_command(auth_cli)
cli.add_command(fine_tune_cli)
cli.add_command(llm_cli)

cli.add_command(anyscale_init)
cli.add_command(anyscale_exec)
cli.add_command(anyscale_help)
cli.add_command(compute_config_cli)
cli.add_command(image_cli)

# Commands to interact with the Anyscale API
cli.add_command(anyscale_api)

cli.add_command(log_cli)
cli.add_command(anyscale_login)
cli.add_command(anyscale_logout)
cli.add_command(machine_cli)
cli.add_command(machine_pool_cli)
cli.add_command(service_account_cli)

ALIASES = {
    "h": anyscale_help,
    "schedules": schedule_cli,
    "jobs": job_cli,
    "services": service_cli,
    "cluster-compute": compute_config_cli,
    "images": image_cli,
    "sa": service_account_cli,
}


def main() -> Any:
    return cli()


if __name__ == "__main__":
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(os.environ.get("ANYSCALE_LOGLEVEL", "WARN"))
    main()
